import inspect
from typing import Any, Callable, Coroutine, Iterator, List, Optional, TypeVar, cast


def iter_methods(
    obj: Any, predicate: Optional[Callable[[Callable[..., Any]], bool]] = None
) -> Iterator[Callable[..., Any]]:
    is_cls = inspect.isclass(obj)
    cls = obj if is_cls else type(obj)

    for name in dir(cls):
        v = getattr(cls, name)
        if inspect.isfunction(v):
            if is_cls:
                m = v
            else:
                m = getattr(obj, name)
                if not inspect.ismethod(m):
                    continue

            if predicate is None or predicate(m):
                yield m


def get_methods(
    instance_or_type: Any, predicate: Optional[Callable[[Callable[..., Any]], bool]] = None
) -> List[Callable[..., Any]]:
    return [m for m in iter_methods(instance_or_type, predicate)]


_lambda_type = type(lambda: 0)
_lambda_name = (lambda: 0).__name__


def is_lambda(v: Any) -> bool:
    return isinstance(v, _lambda_type) and v.__name__ == _lambda_name


_TCallable = TypeVar("_TCallable", bound=Callable[..., Any])


def ensure_coroutine(func: _TCallable) -> Callable[..., Coroutine[Any, Any, Any]]:
    async def wrapper(*fargs: Any, **fkwargs: Any) -> Any:
        return func(*fargs, **fkwargs)

    if inspect.iscoroutinefunction(func) or inspect.iscoroutinefunction(inspect.unwrap(func)):
        return func

    return cast("_TCallable", wrapper)



from __future__ import annotations

import ast
import asyncio
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Type,
    Union,
    cast,
)

from ....utils.async_itertools import async_next
from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import Hover, MarkupContent, MarkupKind, Position
from ...common.text_document import TextDocument
from ..utils.ast import (
    HasTokens,
    Token,
    get_nodes_at_position,
    get_tokens_at_position,
    range_from_node,
    range_from_token,
)
from ..utils.markdownformatter import MarkDownFormatter

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart

_HoverMethod = Callable[[ast.AST, List[ast.AST], TextDocument, Position], Awaitable[Optional[Hover]]]


class RobotHoverProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.hover.collect.add(self.collect)

    def _find_method(self, cls: Type[Any]) -> Optional[_HoverMethod]:
        if cls is ast.AST:
            return None
        method_name = "hover_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast(_HoverMethod, method)
        for base in cls.__bases__:
            method = self._find_method(base)
            if method:
                return cast(_HoverMethod, method)

        return None

    @threaded()
    @language_id("robotframework")
    @_logger.call
    async def collect(self, sender: Any, document: TextDocument, position: Position) -> Optional[Hover]:

        result_nodes = await get_nodes_at_position(await self.parent.documents_cache.get_model(document), position)

        if not result_nodes:
            return None

        result = await self._hover_default(result_nodes, document, position)
        if result:
            return result

        for result_node in reversed(result_nodes):
            method = self._find_method(type(result_node))
            if method is not None:
                result = await method(result_node, result_nodes, document, position)
                if result is not None:
                    return result

        return None

    async def _hover_default(self, nodes: List[ast.AST], document: TextDocument, position: Position) -> Optional[Hover]:
        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        if not nodes:
            return None

        node = nodes[-1]
        if not isinstance(node, HasTokens):
            return None

        tokens = get_tokens_at_position(node, position)

        for token in tokens:
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_variables_from_token(token, namespace, nodes, position)
                    if position in range_from_token(var_token)
                ),
                None,
            )

            # TODO if we found a commandline var, should we look if we found a variable definition and show it in hover?
            if token_and_var is not None:
                var_token, variable = token_and_var

                if variable.has_value or variable.resolvable:
                    try:
                        value = repr(
                            await namespace.imports_manager.resolve_variable(
                                variable.name,
                                str(document.uri.to_path().parent),
                                await namespace.get_resolvable_variables(nodes, position),
                                False,
                            )
                        )
                    except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                        raise
                    except BaseException:
                        value = ""
                else:
                    value = ""

                return Hover(
                    contents=MarkupContent(
                        kind=MarkupKind.MARKDOWN,
                        value=f"({variable.type.value}) {variable.name} {f'= `{value}`' if value else ''}",
                    ),
                    range=range_from_token(var_token),
                )

        return None

    async def hover_IfHeader(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import IfHeader

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        header = cast(IfHeader, node)

        expression_token = header.get_token(RobotToken.ARGUMENT)
        if expression_token is not None and position in range_from_token(expression_token):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        expression_token, namespace, nodes, position
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )
            if token_and_var is not None:
                var_token, variable = token_and_var

                if variable.has_value or variable.resolvable:
                    try:
                        value = await namespace.imports_manager.resolve_variable(
                            variable.name,
                            str(document.uri.to_path().parent),
                            await namespace.get_resolvable_variables(nodes, position),
                            False,
                        )
                    except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                        raise
                    except BaseException:
                        value = ""
                else:
                    value = ""

                return Hover(
                    contents=MarkupContent(
                        kind=MarkupKind.MARKDOWN,
                        value=f"({variable.type.value}) {variable.name} {f'= `{value}`' if value else ''}",
                    ),
                    range=range_from_token(var_token),
                )

        return None

    async def hover_WhileHeader(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import WhileHeader

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        header = cast(WhileHeader, node)

        expression_token = header.get_token(RobotToken.ARGUMENT)
        if expression_token is not None and position in range_from_token(expression_token):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        expression_token, namespace, nodes, position
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )
            if token_and_var is not None:
                var_token, variable = token_and_var

                if variable.has_value or variable.resolvable:
                    try:
                        value = await namespace.imports_manager.resolve_variable(
                            variable.name,
                            str(document.uri.to_path().parent),
                            await namespace.get_resolvable_variables(nodes, position),
                            False,
                        )
                    except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                        raise
                    except BaseException:
                        value = ""
                else:
                    value = ""

                return Hover(
                    contents=MarkupContent(
                        kind=MarkupKind.MARKDOWN,
                        value=f"({variable.type.value}) {variable.name} {f'= `{value}`' if value else ''}",
                    ),
                    range=range_from_token(var_token),
                )

        return None

    async def hover_KeywordCall(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordCall

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordCall, node)
        result = await self.get_keyworddoc_and_token_from_position(
            kw_node.keyword,
            cast(Token, kw_node.get_token(RobotToken.KEYWORD)),
            [cast(Token, t) for t in kw_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    return Hover(
                        contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=lib_entry.library_doc.to_markdown()),
                        range=r,
                    )

            if position in kw_range and keyword_doc is not None and not keyword_doc.is_error_handler:
                return Hover(
                    contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=keyword_doc.to_markdown()),
                    range=kw_range,
                )
        return None

    async def hover_Fixture(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        fixture_node = cast(Fixture, node)

        name_token = cast(Token, fixture_node.get_token(RobotToken.NAME))
        if name_token is None or name_token.value is None or name_token.value in ("", "NONE"):
            return None

        result = await self.get_keyworddoc_and_token_from_position(
            fixture_node.name,
            name_token,
            [cast(Token, t) for t in fixture_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    return Hover(
                        contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=lib_entry.library_doc.to_markdown()),
                        range=r,
                    )

            if position in kw_range and keyword_doc is not None and not keyword_doc.is_error_handler:
                return Hover(
                    contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=keyword_doc.to_markdown()),
                    range=kw_range,
                )
        return None

    async def _hover_Template_or_TestTemplate(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Template, TestTemplate

        template_node = cast(Union[Template, TestTemplate], node)
        if template_node.value:

            keyword_token = cast(RobotToken, template_node.get_token(RobotToken.NAME))
            if keyword_token is None:
                return None

            keyword_token = self.strip_bdd_prefix(keyword_token)

            if position.is_in_range(range_from_token(keyword_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                keyword_doc = await namespace.find_keyword(template_node.value)

                if keyword_doc is not None:

                    lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

                    kw_range = range_from_token(keyword_token)

                    if lib_entry and kw_namespace:
                        r = range_from_token(keyword_token)
                        r.end.character = r.start.character + len(kw_namespace)
                        kw_range.start.character = r.end.character + 1
                        if position in r:
                            return Hover(
                                contents=MarkupContent(
                                    kind=MarkupKind.MARKDOWN, value=lib_entry.library_doc.to_markdown()
                                ),
                                range=r,
                            )
                    if not keyword_doc.is_error_handler:
                        return Hover(
                            contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=keyword_doc.to_markdown()),
                            range=kw_range,
                        )
        return None

    async def hover_TestTemplate(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        return await self._hover_Template_or_TestTemplate(node, nodes, document, position)

    async def hover_Template(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        return await self._hover_Template_or_TestTemplate(node, nodes, document, position)

    async def hover_LibraryImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import LibraryImport

        library_node = cast(LibraryImport, node)
        if library_node.name:

            name_token = cast(RobotToken, library_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            token_range = range_from_token(name_token)
            if position.is_in_range(token_range):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_library_libdoc(
                        library_node.name, library_node.args, library_node.alias
                    )

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_library_import(
                            str(library_node.name), (), str(document.uri.to_path().parent)
                        )

                    if libdoc is None or libdoc.errors:
                        return None

                    return Hover(
                        contents=MarkupContent(
                            kind=MarkupKind.MARKDOWN,
                            value=libdoc.to_markdown(),
                        ),
                        range=token_range,
                    )
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None

    async def hover_ResourceImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import ResourceImport

        resource_node = cast(ResourceImport, node)
        if resource_node.name:

            name_token = cast(RobotToken, resource_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            token_range = range_from_token(name_token)
            if position.is_in_range(token_range):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_resource_libdoc(resource_node.name)

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_resource_import(
                            str(resource_node.name), str(document.uri.to_path().parent)
                        )

                    if libdoc is None or libdoc.errors:
                        return None

                    return Hover(
                        contents=MarkupContent(
                            kind=MarkupKind.MARKDOWN,
                            value=libdoc.to_markdown(),
                        ),
                        range=token_range,
                    )
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None

    async def hover_VariablesImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import VariablesImport

        variables_node = cast(VariablesImport, node)
        if variables_node.name:

            name_token = cast(RobotToken, variables_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            token_range = range_from_token(name_token)
            if position.is_in_range(token_range):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_variables_libdoc(variables_node.name, variables_node.args)

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_variables_import(
                            str(variables_node.name), (), str(document.uri.to_path().parent)
                        )

                    if libdoc is None or libdoc.errors:
                        return None

                    return Hover(
                        contents=MarkupContent(
                            kind=MarkupKind.MARKDOWN,
                            value=libdoc.to_markdown(),
                        ),
                        range=token_range,
                    )
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None

    async def hover_KeywordName(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordName

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordName, node)
        name_token = kw_node.get_token(RobotToken.KEYWORD_NAME)
        if not name_token:
            return None

        result = (await namespace.get_library_doc()).keywords.get(name_token.value, None)

        if result is not None and not result.is_error_handler:
            return Hover(
                contents=MarkupContent(kind=MarkupKind.MARKDOWN, value=result.to_markdown()),
                range=range_from_token(name_token),
            )

        return None

    async def hover_TestCase(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[Hover]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.blocks import TestCase
        from robot.parsing.model.statements import Documentation, Tags

        test_case = cast(TestCase, node)

        if not position.is_in_range(range_from_node(test_case.header)):
            return None

        name_token = cast(RobotToken, test_case.header.get_token(RobotToken.TESTCASE_NAME))
        if name_token is None:
            return None

        doc = next((e for e in test_case.body if isinstance(e, Documentation)), None)
        tags = next((e for e in test_case.body if isinstance(e, Tags)), None)

        txt = f"= Test Case *{test_case.name}* =\n"

        if doc is not None:
            txt += "\n== Documentation ==\n"
            txt += f"\n{doc.value}\n"

        if tags is not None:
            txt += "\n*Tags*: "
            txt += f"{', '.join(tags.values)}\n"

        return Hover(
            contents=MarkupContent(
                kind=MarkupKind.MARKDOWN,
                value=MarkDownFormatter().format(txt),
            ),
            range=range_from_token(name_token),
        )



from __future__ import annotations

import ast
import asyncio
import re
from typing import Any, List, Optional, Union, cast

from ....utils.uri import Uri
from ...common.lsp_types import (
    CodeDescription,
    Diagnostic,
    DiagnosticRelatedInformation,
    DiagnosticSeverity,
    DiagnosticTag,
    Location,
    Position,
    Range,
)
from ...common.text_document import TextDocument
from ..parts.model_helper import ModelHelperMixin
from ..utils.ast import (
    Token,
    is_not_variable_token,
    range_from_node_or_token,
    range_from_token,
)
from ..utils.async_ast import AsyncVisitor
from .library_doc import KeywordDoc, is_embedded_keyword
from .namespace import DIAGNOSTICS_SOURCE_NAME, KeywordFinder, Namespace

EXTRACT_COMMENT_PATTERN = re.compile(r".*(?:^ *|\t+| {2,})#(?P<comment>.*)$")
ROBOTCODE_PATTERN = re.compile(r"(?P<marker>\brobotcode\b)\s*:\s*(?P<rule>\b\w+\b)")


class Analyzer(AsyncVisitor, ModelHelperMixin):
    def __init__(self, model: ast.AST, namespace: Namespace) -> None:
        from robot.parsing.model.statements import Template, TestTemplate

        self.model = model
        self._namespace = namespace
        self.current_testcase_or_keyword_name: Optional[str] = None
        self.finder = KeywordFinder(self._namespace)
        self.test_template: Optional[TestTemplate] = None
        self.template: Optional[Template] = None

    async def run(self) -> List[Diagnostic]:
        self._results: List[Diagnostic] = []

        await self.visit(self.model)
        return self._results

    async def visit(self, node: ast.AST) -> None:
        await super().visit(node)

    @staticmethod
    async def should_ignore(document: Optional[TextDocument], range: Range) -> bool:
        import builtins

        if document is not None:
            lines = await document.get_lines()
            for line_no in builtins.range(range.start.line, range.end.line + 1):
                line = lines[line_no]

                comment = EXTRACT_COMMENT_PATTERN.match(line)
                if comment and comment.group("comment"):
                    for match in ROBOTCODE_PATTERN.finditer(comment.group("comment")):

                        if match.group("rule") == "ignore":
                            return True

        return False

    async def append_diagnostics(
        self,
        range: Range,
        message: str,
        severity: Optional[DiagnosticSeverity] = None,
        code: Union[int, str, None] = None,
        code_description: Optional[CodeDescription] = None,
        source: Optional[str] = None,
        tags: Optional[List[DiagnosticTag]] = None,
        related_information: Optional[List[DiagnosticRelatedInformation]] = None,
        data: Optional[Any] = None,
    ) -> None:

        if await self.should_ignore(self._namespace.document, range):
            return

        self._results.append(
            Diagnostic(range, message, severity, code, code_description, source, tags, related_information, data)
        )

    async def _analyze_keyword_call(
        self,
        keyword: Optional[str],
        node: ast.AST,
        keyword_token: Token,
        argument_tokens: List[Token],
        analyse_run_keywords: bool = True,
    ) -> Optional[KeywordDoc]:
        from robot.parsing.model.statements import Template, TestTemplate

        result: Optional[KeywordDoc] = None

        try:
            if not is_not_variable_token(keyword_token):
                return None

            result = await self.finder.find_keyword(keyword)

            for e in self.finder.diagnostics:
                await self.append_diagnostics(
                    range=range_from_node_or_token(node, self.strip_bdd_prefix(keyword_token)),
                    message=e.message,
                    severity=e.severity,
                    source=DIAGNOSTICS_SOURCE_NAME,
                    code=e.code,
                )

            if result is not None:
                if result.errors:
                    await self.append_diagnostics(
                        range=range_from_node_or_token(node, self.strip_bdd_prefix(keyword_token)),
                        message="Keyword definition contains errors.",
                        severity=DiagnosticSeverity.ERROR,
                        source=DIAGNOSTICS_SOURCE_NAME,
                        related_information=[
                            DiagnosticRelatedInformation(
                                location=Location(
                                    uri=str(
                                        Uri.from_path(
                                            err.source
                                            if err.source is not None
                                            else result.source
                                            if result.source is not None
                                            else "/<unknown>"
                                        )
                                    ),
                                    range=Range(
                                        start=Position(
                                            line=err.line_no - 1
                                            if err.line_no is not None
                                            else result.line_no
                                            if result.line_no >= 0
                                            else 0,
                                            character=0,
                                        ),
                                        end=Position(
                                            line=err.line_no - 1
                                            if err.line_no is not None
                                            else result.line_no
                                            if result.line_no >= 0
                                            else 0,
                                            character=0,
                                        ),
                                    ),
                                ),
                                message=err.message,
                            )
                            for err in result.errors
                        ],
                    )

                if result.is_deprecated:
                    await self.append_diagnostics(
                        range=range_from_node_or_token(node, self.strip_bdd_prefix(keyword_token)),
                        message=f"Keyword '{result.name}' is deprecated"
                        f"{f': {result.deprecated_message}' if result.deprecated_message else ''}.",
                        severity=DiagnosticSeverity.HINT,
                        source=DIAGNOSTICS_SOURCE_NAME,
                        tags=[DiagnosticTag.Deprecated],
                    )
                if result.is_error_handler:
                    await self.append_diagnostics(
                        range=range_from_node_or_token(node, self.strip_bdd_prefix(keyword_token)),
                        message=f"Keyword definition contains errors: {result.error_handler_message}",
                        severity=DiagnosticSeverity.ERROR,
                        source=DIAGNOSTICS_SOURCE_NAME,
                    )
                if result.is_reserved():
                    await self.append_diagnostics(
                        range=range_from_node_or_token(node, self.strip_bdd_prefix(keyword_token)),
                        message=f"'{result.name}' is a reserved keyword.",
                        severity=DiagnosticSeverity.ERROR,
                        source=DIAGNOSTICS_SOURCE_NAME,
                    )

                if not isinstance(node, (Template, TestTemplate)):
                    try:
                        if result.arguments is not None:
                            result.arguments.resolve(
                                [v.value for v in argument_tokens],
                                None,
                                resolve_variables_until=result.args_to_process,
                                resolve_named=not result.is_any_run_keyword(),
                            )
                    except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                        raise
                    except BaseException as e:
                        await self.append_diagnostics(
                            range=Range(
                                start=range_from_token(self.strip_bdd_prefix(keyword_token)).start,
                                end=range_from_token(argument_tokens[-1]).end
                                if argument_tokens
                                else range_from_token(keyword_token).end,
                            ),
                            message=str(e),
                            severity=DiagnosticSeverity.ERROR,
                            source=DIAGNOSTICS_SOURCE_NAME,
                            code=type(e).__qualname__,
                        )

        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            await self.append_diagnostics(
                range=range_from_node_or_token(node, keyword_token),
                message=str(e),
                severity=DiagnosticSeverity.ERROR,
                source=DIAGNOSTICS_SOURCE_NAME,
                code=type(e).__qualname__,
            )

        if result is not None and analyse_run_keywords:
            await self._analyse_run_keyword(result, node, argument_tokens)

        return result

    async def _analyse_run_keyword(
        self, keyword_doc: Optional[KeywordDoc], node: ast.AST, argument_tokens: List[Token]
    ) -> List[Token]:
        from robot.utils.escaping import unescape

        if keyword_doc is None or not keyword_doc.is_any_run_keyword():
            return argument_tokens

        if keyword_doc.is_run_keyword() and len(argument_tokens) > 0 and is_not_variable_token(argument_tokens[0]):
            await self._analyze_keyword_call(
                unescape(argument_tokens[0].value), node, argument_tokens[0], argument_tokens[1:]
            )

            return argument_tokens[1:]
        elif (
            keyword_doc.is_run_keyword_with_condition()
            and len(argument_tokens) > (cond_count := keyword_doc.run_keyword_condition_count())
            and is_not_variable_token(argument_tokens[cond_count])
        ):
            await self._analyze_keyword_call(
                unescape(argument_tokens[cond_count].value),
                node,
                argument_tokens[cond_count],
                argument_tokens[cond_count + 1 :],
            )
            return argument_tokens[cond_count + 1 :]
        elif keyword_doc.is_run_keywords():
            has_and = False
            while argument_tokens:

                t = argument_tokens[0]
                argument_tokens = argument_tokens[1:]
                if t.value == "AND":
                    await self.append_diagnostics(
                        range=range_from_token(t),
                        message=f"Incorrect use of {t.value}",
                        severity=DiagnosticSeverity.ERROR,
                        source=DIAGNOSTICS_SOURCE_NAME,
                    )
                    continue

                if not is_not_variable_token(t):
                    continue

                and_token = next((e for e in argument_tokens if e.value == "AND"), None)
                args = []
                if and_token is not None:
                    args = argument_tokens[: argument_tokens.index(and_token)]
                    argument_tokens = argument_tokens[argument_tokens.index(and_token) + 1 :]
                    has_and = True
                elif has_and:
                    args = argument_tokens
                    argument_tokens = []

                await self._analyze_keyword_call(unescape(t.value), node, t, args)

            return []

        elif keyword_doc.is_run_keyword_if() and len(argument_tokens) > 1:

            def skip_args() -> List[Token]:
                nonlocal argument_tokens
                result = []
                while argument_tokens:
                    if argument_tokens[0].value in ["ELSE", "ELSE IF"]:
                        break
                    if argument_tokens:
                        result.append(argument_tokens[0])
                    argument_tokens = argument_tokens[1:]

                return result

            result = await self.finder.find_keyword(argument_tokens[1].value)

            if result is not None and result.is_any_run_keyword():
                argument_tokens = argument_tokens[2:]

                argument_tokens = await self._analyse_run_keyword(result, node, argument_tokens)
            else:
                kwt = argument_tokens[1]
                argument_tokens = argument_tokens[2:]

                args = skip_args()

                if is_not_variable_token(kwt):
                    await self._analyze_keyword_call(
                        unescape(kwt.value),
                        node,
                        kwt,
                        args,
                        analyse_run_keywords=False,
                    )

            while argument_tokens:
                if argument_tokens[0].value == "ELSE" and len(argument_tokens) > 1:

                    result = await self._analyze_keyword_call(
                        unescape(argument_tokens[1].value),
                        node,
                        argument_tokens[1],
                        argument_tokens[2:],
                        analyse_run_keywords=False,
                    )

                    argument_tokens = argument_tokens[2:]

                    if result is not None and result.is_any_run_keyword():
                        argument_tokens = await self._analyse_run_keyword(result, node, argument_tokens)

                    skip_args()

                    break
                elif argument_tokens[0].value == "ELSE IF" and len(argument_tokens) > 2:

                    result = await self._analyze_keyword_call(
                        unescape(argument_tokens[2].value),
                        node,
                        argument_tokens[2],
                        argument_tokens[3:],
                        analyse_run_keywords=False,
                    )

                    argument_tokens = argument_tokens[3:]

                    if result is not None and result.is_any_run_keyword():
                        argument_tokens = await self._analyse_run_keyword(result, node, argument_tokens)

                    skip_args()
                else:
                    break

        return argument_tokens

    async def visit_Fixture(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        value = cast(Fixture, node)
        keyword_token = cast(Token, value.get_token(RobotToken.NAME))

        # TODO: calculate possible variables in NAME

        if (
            keyword_token is not None
            and is_not_variable_token(keyword_token)
            and keyword_token.value is not None
            and keyword_token.value.upper() not in ("", "NONE")
        ):
            await self._analyze_keyword_call(
                value.name, value, keyword_token, [cast(Token, e) for e in value.get_tokens(RobotToken.ARGUMENT)]
            )

        await self.generic_visit(node)

    async def visit_TestTemplate(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import TestTemplate

        value = cast(TestTemplate, node)
        keyword_token = cast(Token, value.get_token(RobotToken.NAME))

        # TODO: calculate possible variables in NAME

        if keyword_token is not None and is_not_variable_token(keyword_token):
            await self._analyze_keyword_call(value.value, value, keyword_token, [])

        self.test_template = value
        await self.generic_visit(node)

    async def visit_Template(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Template

        value = cast(Template, node)
        keyword_token = cast(Token, value.get_token(RobotToken.NAME))

        # TODO: calculate possible variables in NAME

        if keyword_token is not None and is_not_variable_token(keyword_token):
            await self._analyze_keyword_call(value.value, value, keyword_token, [])

        await self.generic_visit(node)

    async def visit_KeywordCall(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordCall

        value = cast(KeywordCall, node)
        keyword_token = cast(RobotToken, value.get_token(RobotToken.KEYWORD))

        if value.assign and not value.keyword:
            await self.append_diagnostics(
                range=range_from_node_or_token(value, value.get_token(RobotToken.ASSIGN)),
                message="Keyword name cannot be empty.",
                severity=DiagnosticSeverity.ERROR,
                source=DIAGNOSTICS_SOURCE_NAME,
                code="KeywordError",
            )
        else:
            await self._analyze_keyword_call(
                value.keyword, value, keyword_token, [cast(Token, e) for e in value.get_tokens(RobotToken.ARGUMENT)]
            )

        if not self.current_testcase_or_keyword_name:
            await self.append_diagnostics(
                range=range_from_node_or_token(value, value.get_token(RobotToken.ASSIGN)),
                message="Code is unreachable.",
                severity=DiagnosticSeverity.HINT,
                source=DIAGNOSTICS_SOURCE_NAME,
                tags=[DiagnosticTag.Unnecessary],
            )

        await self.generic_visit(node)

    async def visit_TestCase(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.blocks import TestCase
        from robot.parsing.model.statements import TestCaseName

        testcase = cast(TestCase, node)

        if not testcase.name:
            name_token = cast(TestCaseName, testcase.header).get_token(RobotToken.TESTCASE_NAME)
            await self.append_diagnostics(
                range=range_from_node_or_token(testcase, name_token),
                message="Test case name cannot be empty.",
                severity=DiagnosticSeverity.ERROR,
                source=DIAGNOSTICS_SOURCE_NAME,
                code="KeywordError",
            )

        self.current_testcase_or_keyword_name = testcase.name
        try:
            await self.generic_visit(node)
        finally:
            self.current_testcase_or_keyword_name = None

    async def visit_Keyword(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.blocks import Keyword
        from robot.parsing.model.statements import Arguments, KeywordName

        keyword = cast(Keyword, node)

        if keyword.name:

            name_token = cast(KeywordName, keyword.header).get_token(RobotToken.KEYWORD_NAME)
            if is_embedded_keyword(keyword.name) and any(
                isinstance(v, Arguments) and len(v.values) > 0 for v in keyword.body
            ):
                await self.append_diagnostics(
                    range=range_from_node_or_token(keyword, name_token),
                    message="Keyword cannot have both normal and embedded arguments.",
                    severity=DiagnosticSeverity.ERROR,
                    source=DIAGNOSTICS_SOURCE_NAME,
                    code="KeywordError",
                )
        else:
            name_token = cast(KeywordName, keyword.header).get_token(RobotToken.KEYWORD_NAME)
            await self.append_diagnostics(
                range=range_from_node_or_token(keyword, name_token),
                message="Keyword name cannot be empty.",
                severity=DiagnosticSeverity.ERROR,
                source=DIAGNOSTICS_SOURCE_NAME,
                code="KeywordError",
            )

        self.current_testcase_or_keyword_name = keyword.name
        try:
            await self.generic_visit(node)
        finally:
            self.current_testcase_or_keyword_name = None



from __future__ import annotations

import asyncio
import itertools
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, cast

from ....utils.async_tools import (
    Lock,
    async_tasking_event_iterator,
    check_canceled,
    create_sub_task,
    run_coroutine_in_thread,
)
from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ..decorators import language_id, language_id_filter
from ..lsp_types import Diagnostic, DocumentUri, PublishDiagnosticsParams
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart

__all__ = ["DiagnosticsProtocolPart", "DiagnosticsResult"]

DIAGNOSTICS_DEBOUNCE = 0.75


class PublishDiagnosticsEntry:
    _logger = LoggingDescriptor()

    def __init__(
        self,
        uri: Uri,
        version: Optional[int],
        factory: Callable[..., asyncio.Future[Any]],
        done_callback: Callable[[PublishDiagnosticsEntry], Any],
    ) -> None:

        self.uri = uri
        self.version = version

        self._factory = factory
        self.done_callback = done_callback

        self._future: Optional[asyncio.Future[Any]] = None

        self.done = False

        def _done(t: asyncio.Future[Any]) -> None:
            self.done = True
            self.done_callback(self)

        self._future = create_sub_task(self._wait_and_run())
        self._future.add_done_callback(_done)

    async def _wait_and_run(self) -> None:
        await asyncio.sleep(DIAGNOSTICS_DEBOUNCE)
        await self._factory()

    def __del__(self) -> None:
        if not self.done:
            try:
                if asyncio.get_running_loop():
                    create_sub_task(self.cancel())
            except RuntimeError:
                pass

    @property
    def future(self) -> Optional[asyncio.Future[Any]]:
        return self._future

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{type(self)}(document={repr(self.uri)}, task={repr(self.future)}, done={self.done})"

    @_logger.call
    async def cancel(self) -> None:
        if self.future is None:
            return

        self.done = True

        if not self.future.done() and not self.future.cancelled():
            self.future.cancel()
            try:
                await self.future
            except (asyncio.CancelledError):
                pass
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as ex:
                self._logger.exception(ex)
                raise
            finally:
                self._future = None


@dataclass
class DiagnosticsResult:
    key: Any
    diagnostics: Optional[List[Diagnostic]] = None


class DiagnosticsProtocolPart(LanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, protocol: LanguageServerProtocol) -> None:
        super().__init__(protocol)

        self._running_diagnostics: Dict[Uri, PublishDiagnosticsEntry] = {}
        self._tasks_lock = Lock()

        self.parent.on_connection_lost.add(self.on_connection_lost)
        self.parent.on_shutdown.add(self.on_shutdown)
        self.parent.documents.did_open.add(self.on_did_open)
        self.parent.documents.did_change.add(self.on_did_change)
        self.parent.documents.did_close.add(self.on_did_close)
        self.parent.documents.did_save.add(self.on_did_save)

    @async_tasking_event_iterator
    async def collect(sender, document: TextDocument) -> DiagnosticsResult:  # NOSONAR
        ...

    @_logger.call
    async def on_connection_lost(self, sender: Any, exc: Optional[BaseException]) -> None:
        await self._cancel_all_tasks()

    @_logger.call
    async def on_shutdown(self, sender: Any) -> None:
        await self._cancel_all_tasks()

    def __del__(self) -> None:
        if len(self._running_diagnostics) > 0:
            self._logger.warning("there are running tasks")

    @_logger.call
    async def _cancel_all_tasks(self) -> None:
        tasks_copy = None
        async with self._tasks_lock:
            tasks_copy = self._running_diagnostics.copy()
            self._running_diagnostics = {}
        if tasks_copy is not None:
            for v in tasks_copy.values():
                await self._cancel_entry(v)

    @_logger.call(condition=lambda self, entry: entry is not None)
    async def _cancel_entry(self, entry: Optional[PublishDiagnosticsEntry]) -> None:
        if entry is None:
            return
        if not entry.done:
            await entry.cancel()

    @language_id("robotframework")
    @_logger.call
    async def on_did_open(self, sender: Any, document: TextDocument) -> None:
        create_sub_task(self.start_publish_diagnostics_task(document))

    @language_id("robotframework")
    @_logger.call
    async def on_did_save(self, sender: Any, document: TextDocument) -> None:
        create_sub_task(self.start_publish_diagnostics_task(document))

    @language_id("robotframework")
    @_logger.call
    async def on_did_close(self, sender: Any, document: TextDocument) -> None:
        try:
            await self._cancel_entry(self._running_diagnostics.get(document.uri, None))
        finally:
            self.parent.send_notification(
                "textDocument/publishDiagnostics",
                PublishDiagnosticsParams(
                    uri=document.document_uri,
                    version=document._version,
                    diagnostics=[],
                ),
            )

    @_logger.call
    async def on_did_change(self, sender: Any, document: TextDocument) -> None:
        create_sub_task(self.start_publish_diagnostics_task(document))

    @_logger.call
    def _delete_entry(self, e: PublishDiagnosticsEntry) -> None:
        if e.uri in self._running_diagnostics and self._running_diagnostics[e.uri] == e:
            self._running_diagnostics.pop(e.uri, None)

    @_logger.call
    async def start_publish_diagnostics_task(self, document: TextDocument) -> None:
        async with self._tasks_lock:
            entry = self._running_diagnostics.get(document.uri, None)

            if entry is not None and entry.version == document.version:
                return

            await self._cancel_entry(entry)

            self._running_diagnostics[document.uri] = PublishDiagnosticsEntry(
                document.uri,
                document.version,
                lambda: run_coroutine_in_thread(self.publish_diagnostics, document.document_uri),
                self._delete_entry,
            )

    @_logger.call
    async def publish_diagnostics(self, document_uri: DocumentUri) -> None:
        document = await self.parent.documents.get(document_uri)
        if document is None:
            return

        diagnostics: Dict[Any, List[Diagnostic]] = document.get_data(self, {})

        collected_keys: List[Any] = []

        async for result_any in self.collect(
            self,
            document,
            callback_filter=language_id_filter(document),
            return_exceptions=True,
        ):
            await check_canceled()

            result = cast(DiagnosticsResult, result_any)

            if isinstance(result, BaseException):
                if not isinstance(result, asyncio.CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:

                diagnostics[result.key] = result.diagnostics if result.diagnostics else []
                collected_keys.append(result.key)

                asyncio.get_event_loop().call_soon(
                    self.parent.send_notification,
                    "textDocument/publishDiagnostics",
                    PublishDiagnosticsParams(
                        uri=document.document_uri,
                        version=document._version,
                        diagnostics=[e for e in itertools.chain(*diagnostics.values())],
                    ),
                )

        for k in set(diagnostics.keys()) - set(collected_keys):
            diagnostics.pop(k)

        document.set_data(self, diagnostics)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union, cast

from .dap_types import Event, Model
from .debugger import Debugger


@dataclass
class RobotExecutionEventBody(Model):
    type: str
    id: str
    attributes: Optional[Dict[str, Any]] = None
    failed_keywords: Optional[List[Dict[str, Any]]] = None


class ListenerV2:
    ROBOT_LISTENER_API_VERSION = "2"

    def __init__(self, no_debug: bool = False) -> None:
        self.no_debug = no_debug
        self.debug = not no_debug
        self.failed_keywords: Optional[List[Dict[str, Any]]] = None
        self.last_fail_message: Optional[str] = None

    def start_suite(self, name: str, attributes: Dict[str, Any]) -> None:
        Debugger.instance().send_event(
            self,
            Event(
                event="robotStarted",
                body=RobotExecutionEventBody(
                    type="suite",
                    id=f"{attributes.get('source', '')};{attributes.get('longname', '')}",
                    attributes=dict(attributes),
                ),
            ),
        )

        Debugger.instance().start_output_group(name, attributes, "SUITE")

        Debugger.instance().start_suite(name, attributes)

    def end_suite(self, name: str, attributes: Dict[str, Any]) -> None:

        Debugger.instance().end_suite(name, attributes)

        Debugger.instance().end_output_group(name, attributes)

        Debugger.instance().send_event(
            self,
            Event(
                event="robotEnded",
                body=RobotExecutionEventBody(
                    type="suite",
                    attributes=dict(attributes),
                    id=f"{attributes.get('source', '')};{attributes.get('longname', '')}",
                    failed_keywords=self.failed_keywords,
                ),
            ),
        )

    def start_test(self, name: str, attributes: Dict[str, Any]) -> None:
        self.failed_keywords = None

        Debugger.instance().send_event(
            self,
            Event(
                event="robotStarted",
                body=RobotExecutionEventBody(
                    type="test",
                    id=f"{attributes.get('source', '')};{attributes.get('longname', '')};"
                    f"{attributes.get('lineno', 0)}",
                    attributes=dict(attributes),
                ),
            ),
        )

        Debugger.instance().start_output_group(name, attributes, "TEST")

        Debugger.instance().start_test(name, attributes)

    def end_test(self, name: str, attributes: Dict[str, Any]) -> None:
        Debugger.instance().end_test(name, attributes)

        Debugger.instance().end_output_group(name, attributes)

        Debugger.instance().send_event(
            self,
            Event(
                event="robotEnded",
                body=RobotExecutionEventBody(
                    type="test",
                    id=f"{attributes.get('source', '')};{attributes.get('longname', '')};"
                    f"{attributes.get('lineno', 0)}",
                    attributes=dict(attributes),
                    failed_keywords=self.failed_keywords,
                ),
            ),
        )

        self.failed_keywords = None

    def start_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        if attributes["type"] in ["KEYWORD", "SETUP", "TEARDOWN"]:
            Debugger.instance().start_output_group(
                f"{name}({', '.join(repr(v) for v in attributes.get('args', []))})",
                attributes,
                attributes.get("type", None),
            )

        Debugger.instance().start_keyword(name, attributes)

    def end_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        Debugger.instance().end_keyword(name, attributes)

        if attributes["type"] in ["KEYWORD", "SETUP", "TEARDOWN"]:
            Debugger.instance().end_output_group(name, attributes)

            if attributes["status"] == "FAIL" and attributes.get("source", None):
                if self.failed_keywords is None:
                    self.failed_keywords = []

                self.failed_keywords.insert(0, {"message": self.last_fail_message, **attributes})

    def log_message(self, message: Dict[str, Any]) -> None:
        current_frame = Debugger.instance().full_stack_frames[0] if Debugger.instance().full_stack_frames else None

        if message["level"] == "FAIL":
            self.last_fail_message = message["message"]
            Debugger.instance().last_fail_message = self.last_fail_message

        source = current_frame.source if current_frame else None
        line = current_frame.line if current_frame else None
        column = current_frame.column if current_frame else None

        name = next((e.name for e in Debugger.instance().full_stack_frames if e.type in ["SUITE", "TEST"]), None)

        Debugger.instance().send_event(
            self,
            Event(
                event="robotLog",
                body={"itemId": name, "source": source, "lineno": line, "column": column, **dict(message)},
            ),
        )

        Debugger.instance().log_message(message)

    def message(self, message: Dict[str, Any]) -> None:
        Debugger.instance().message(message)

    def library_import(self, name: str, attributes: Dict[str, Any]) -> None:
        pass

    def resource_import(self, name: str, attributes: Dict[str, Any]) -> None:
        pass

    def variables_import(self, name: str, attributes: Dict[str, Any]) -> None:
        pass

    def output_file(self, path: str) -> None:
        Debugger.instance().robot_output_file = path

    def log_file(self, path: str) -> None:
        Debugger.instance().robot_log_file = path

    def report_file(self, path: str) -> None:
        Debugger.instance().robot_report_file = path

    def xunit_file(self, path: str) -> None:
        pass

    def debug_file(self, path: str) -> None:
        pass

    def close(self) -> None:
        pass


class ListenerV3:
    ROBOT_LISTENER_API_VERSION = "3"

    def __init__(self) -> None:
        self._event_sended = False

    def start_suite(self, data: Any, result: Any) -> None:
        from robot.running import TestCase, TestSuite

        def enqueue(item: Union[TestSuite, TestCase]) -> Iterator[str]:
            if isinstance(item, TestSuite):
                for s in item.suites:
                    yield from enqueue(s)
                for s in item.tests:
                    yield from enqueue(s)

                yield f"{Path(item.source).resolve()};{item.longname}"
                return

            yield f"{Path(item.source).resolve()};{item.longname};{item.lineno}"

        if self._event_sended:
            return

        items = list(reversed([i for i in enqueue(cast(TestSuite, data))]))

        Debugger.instance().send_event(
            self,
            Event(
                event="robotEnqueued",
                body={"items": items},
            ),
        )

        self._event_sended = True

    def end_suite(self, data: Any, result: Any) -> None:
        pass

    def start_test(self, data: Any, result: Any) -> None:
        pass

    def end_test(self, data: Any, result: Any) -> None:
        pass



from __future__ import annotations

import functools
import itertools
import re
from abc import ABC, abstractmethod
from typing import Any, Callable, Iterator, List, Optional, Tuple


class Formatter(ABC):
    _strip_lines = True

    def __init__(self) -> None:
        self._lines: List[str] = []

    def handles(self, line: str) -> bool:
        return self._handles(line.strip() if self._strip_lines else line)

    @abstractmethod
    def _handles(self, line: str) -> bool:
        ...

    def add(self, line: str) -> None:
        self._lines.append(line.strip() if self._strip_lines else line)

    def end(self) -> str:
        result = self.format(self._lines)
        self._lines = []
        return result

    @abstractmethod
    def format(self, lines: List[str]) -> str:
        ...


class MarkDownFormatter:
    def __init__(self) -> None:
        self._results: List[str] = []
        self._formatters: List[Formatter] = [
            TableFormatter(),
            PreformattedFormatter(),
            ListFormatter(),
            HeaderFormatter(),
            RulerFormatter(),
        ]
        self._formatters.append(ParagraphFormatter(self._formatters[:]))
        self._current: Optional[Formatter] = None

    def format(self, text: str) -> str:
        for line in text.splitlines():
            self._process_line(line)
        self._end_current()
        return "\n".join(self._results)

    def _process_line(self, line: str) -> None:
        if not line.strip():
            self._end_current()
        elif self._current and self._current.handles(line):
            self._current.add(line)
        else:
            self._end_current()
            self._current = self._find_formatter(line)
            if self._current is not None:
                self._current.add(line)

    def _end_current(self) -> None:
        if self._current:
            self._results.append(self._current.end())
            self._current = None

    def _find_formatter(self, line: str) -> Optional[Formatter]:
        for formatter in self._formatters:
            if formatter.handles(line):
                return formatter
        return None


class SingleLineFormatter(Formatter):
    def _handles(self, line: str) -> bool:
        return bool(not self._lines and self.match(line))

    @abstractmethod
    def match(self, line: str) -> Optional[re.Match[str]]:
        ...

    def format(self, lines: List[str]) -> str:
        return self.format_line(lines[0])

    @abstractmethod
    def format_line(self, line: str) -> str:
        ...


class HeaderFormatter(SingleLineFormatter):
    _regex = re.compile(r"^(={1,5})\s+(\S.*?)\s+\1$")

    def match(self, line: str) -> Optional[re.Match[str]]:
        return self._regex.match(line)

    def format_line(self, line: str) -> str:
        m = self.match(line)
        if m is not None:
            level, text = m.groups()

            return "%s %s\n" % ("#" * (len(level) + 1), text)
        return ""


class LinkFormatter:
    _image_exts = (".jpg", ".jpeg", ".png", ".gif", ".bmp", ".svg")
    _link = re.compile(r"\[(.+?\|.*?)\]")
    _url = re.compile(
        r"""
((^|\ ) ["'(\[{]*)           # begin of line or space and opt. any char "'([{
([a-z][\w+-.]*://[^\s|]+?)   # url
(?=[)\]}"'.,!?:;|]* ($|\ ))  # opt. any char )]}"'.,!?:;| and eol or space
""",
        re.VERBOSE | re.MULTILINE | re.IGNORECASE,
    )

    def format_url(self, text: str) -> str:
        return self._format_url(text, format_as_image=False)

    def _format_url(self, text: str, format_as_image: bool = True) -> str:
        if "://" not in text:
            return text
        return self._url.sub(functools.partial(self._replace_url, format_as_image), text)

    def _replace_url(self, format_as_image: bool, match: re.Match[str]) -> str:
        pre = match.group(1)
        url = match.group(3)
        if format_as_image and self._is_image(url):
            return pre + self._get_image(url)
        return pre + self._get_link(url)

    def _get_image(self, src: str, title: Optional[str] = None) -> str:
        return f"![{title or src}]({src})"

    def _get_link(self, href: str, content: Optional[str] = None) -> str:
        return f"[{content or href}]({href})"

    def _quot(self, attr: str) -> str:
        return attr if '"' not in attr else attr.replace('"', "&quot;")

    def format_link(self, text: str) -> str:
        # 2nd, 4th, etc. token contains link, others surrounding content
        tokens = self._link.split(text)

        formatters: Iterator[Callable[[str], Any]] = itertools.cycle((self._format_url, self._format_link))
        return "".join(f(t) for f, t in zip(formatters, tokens))

    def _format_link(self, text: str) -> str:
        link, content = [t.strip() for t in text.split("|", 1)]
        if self._is_image(content):
            content = self._get_image(content, link)
        elif self._is_image(link):
            return self._get_image(link, content)
        return self._get_link(link, content)

    def remove_link(self, text: str) -> str:
        # 2nd, 4th, etc. token contains link, others surrounding content
        tokens = self._link.split(text)
        if len(tokens) > 1:
            formatters: Iterator[Callable[[str], Any]] = itertools.cycle([self._remove_link])
            return "".join(f(t) for f, t in zip(formatters, tokens))
        return text

    def _remove_link(self, text: str) -> str:
        if "|" not in text:
            return text

        link, content = [t.strip() for t in text.split("|", 1)]
        if self._is_image(content):
            content = self._get_image(content, link)

        return content

    def _is_image(self, text: str) -> bool:
        return text.startswith("data:image/") or text.lower().endswith(self._image_exts)


class LineFormatter:
    _bold = re.compile(
        r"""
(                         # prefix (group 1)
  (^|\ )                  # begin of line or space
  ["'(]* _?               # optionally any char "'( and optional begin of italic
)                         #
\*                        # start of bold
([^\ ].*?)                # no space and then anything (group 3)
\*                        # end of bold
(?=                       # start of postfix (non-capturing group)
  _? ["').,!?:;]*         # optional end of italic and any char "').,!?:;
  ($|\ )                  # end of line or space
)
""",
        re.VERBOSE,
    )
    _italic = re.compile(
        r"""
( (^|\ ) ["'(]* )          # begin of line or space and opt. any char "'(
_                          # start of italic
([^\ _].*?)                # no space or underline and then anything
_                          # end of italic
(?= ["').,!?:;]* ($|\ ) )  # opt. any char "').,!?:; and end of line or space
""",
        re.VERBOSE,
    )
    _code = re.compile(
        r"""
( (^|\ ) ["'(]* )          # same as above with _ changed to ``
``
([^\ `].*?)
``
(?= ["').,!?:;]* ($|\ ) )
""",
        re.VERBOSE,
    )

    def __init__(self) -> None:
        super().__init__()

        self._formatters: List[Tuple[str, Callable[[str], str]]] = [
            ("<", self._quote_lower_then),
            ("#", self._quote_hash),
            ("*", self._format_bold),
            ("_", self._format_italic),
            ("``", self._format_code),
            ("", functools.partial(LinkFormatter().format_link)),
        ]

    def format(self, line: str) -> str:
        for marker, formatter in self._formatters:
            if marker in line:
                line = formatter(line)
        return line

    def _quote_lower_then(self, line: str) -> str:
        return line.replace("<", "\\<")

    def _quote_hash(self, line: str) -> str:
        return line.replace("#", "\\#")

    def _format_bold(self, line: str) -> str:
        return self._bold.sub("\\1**\\3**", line)

    def _format_italic(self, line: str) -> str:
        return self._italic.sub("\\1*\\3*", line)

    def _format_code(self, line: str) -> str:
        return self._code.sub("\\1`\\3`", line)


class PreformattedFormatter(Formatter):
    _format_line = functools.partial(LineFormatter().format)

    def _handles(self, line: str) -> bool:
        return line.startswith("| ") or line == "|"

    def format(self, lines: List[str]) -> str:
        lines = [LinkFormatter().remove_link(line[2:]) for line in lines]
        return "```text\n" + "\n".join(lines) + "\n```\n"


class ParagraphFormatter(Formatter):
    _format_line = functools.partial(LineFormatter().format)

    def __init__(self, other_formatters: List[Formatter]) -> None:
        super().__init__()
        self._other_formatters = other_formatters

    def _handles(self, line: str) -> bool:
        return not any(other.handles(line) for other in self._other_formatters)

    def format(self, lines: List[str]) -> str:
        return self._format_line(" ".join(lines)) + "\n\n"


class ListFormatter(Formatter):
    _strip_lines = False
    _format_item = functools.partial(LineFormatter().format)

    def _handles(self, line: str) -> bool:
        return bool(line.strip().startswith("- ") or line.startswith(" ") and self._lines)

    def format(self, lines: List[str]) -> str:
        items = ["- %s" % self._format_item(line) for line in self._combine_lines(lines)]
        return "\n".join(items) + "\n\n"

    def _combine_lines(self, lines: List[str]) -> Iterator[str]:
        current = []
        for line in lines:
            line = line.strip()
            if not line.startswith("- "):
                current.append(line)
                continue
            if current:
                yield " ".join(current)
            current = [line[2:].strip()]
        yield " ".join(current)


class RulerFormatter(SingleLineFormatter):
    regex = re.compile("^-{3,}$")

    def match(self, line: str) -> Optional[re.Match[str]]:
        return self.regex.match(line)

    def format_line(self, line: str) -> str:
        return "---"


class TableFormatter(Formatter):
    _table_line = re.compile(r"^\| (.* |)\|$")
    _line_splitter = re.compile(r" \|(?= )")
    _format_cell_content = functools.partial(LineFormatter().format)

    def _handles(self, line: str) -> bool:
        return self._table_line.match(line) is not None

    def format(self, lines: List[str]) -> str:
        return self._format_table([self._split_to_cells(line) for line in lines])

    def _split_to_cells(self, line: str) -> List[str]:
        return [cell.strip() for cell in self._line_splitter.split(line[1:-1])]

    def _format_table(self, rows: List[List[str]]) -> str:
        table = []

        max_columns = max(len(row) for row in rows)

        try:
            header_rows = [list(next(row for row in rows if any(cell for cell in row if cell.startswith("="))))]
        except StopIteration:
            header_rows = [[]]

        body_rows = [row for row in rows if row not in header_rows]

        for row in header_rows or [[]]:
            row += [""] * (max_columns - len(row))
            table.append(f'|{"|".join(self._format_cell(cell) for cell in row)}|')

        row_ = [" :--- "] * max_columns
        table.append(f'|{"|".join(row_)}|')

        for row in body_rows:
            row += [""] * (max_columns - len(row))
            table.append(f'|{"|".join(self._format_cell(cell) for cell in row)}|')

        return "\n".join(table) + "\n\n"

    def _format_cell(self, content: str) -> str:
        if content.startswith("=") and content.endswith("="):
            content = content[1:-1]

        return f" {self._format_cell_content(content).strip()} "



from __future__ import annotations

import ast
import itertools
from typing import TYPE_CHECKING, Any, List, Optional, Union, cast

from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import DocumentSymbol, SymbolInformation, SymbolKind
from ...common.text_document import TextDocument
from ..utils.ast import Token, range_from_node, range_from_token, tokenize_variables

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .protocol_part import RobotLanguageServerProtocolPart


class RobotDocumentSymbolsProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.document_symbols.collect.add(self.collect)

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect(
        self, sender: Any, document: TextDocument
    ) -> Optional[Union[List[DocumentSymbol], List[SymbolInformation], None]]:

        from ..utils.async_ast import AsyncVisitor

        class Visitor(AsyncVisitor):
            def __init__(self, parent: RobotDocumentSymbolsProtocolPart) -> None:
                super().__init__()
                self.parent = parent

                self.result: List[DocumentSymbol] = []
                self.current_symbol: Optional[DocumentSymbol] = None

            async def generic_visit_current_symbol(self, node: ast.AST, symbol: DocumentSymbol) -> None:
                old = self.current_symbol
                self.current_symbol = symbol
                try:
                    await self.generic_visit(node)
                finally:
                    self.current_symbol = old

            @classmethod
            async def find_from(
                cls, model: ast.AST, parent: RobotDocumentSymbolsProtocolPart
            ) -> Optional[List[DocumentSymbol]]:
                finder = cls(parent)

                await finder.visit(model)

                return finder.result if finder.result else None

            async def visit_Section(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import Section
                from robot.parsing.model.statements import SectionHeader

                section = cast(Section, node)
                if section.header is None:
                    return

                header = cast(SectionHeader, section.header)
                if header.name is None:
                    return

                r = range_from_node(section)
                symbol = DocumentSymbol(
                    name=header.name.replace("*", "").strip(),
                    kind=SymbolKind.NAMESPACE,
                    range=r,
                    selection_range=r,
                    children=[],
                )

                self.result.append(symbol)

                await self.generic_visit_current_symbol(node, symbol)

            async def visit_TestCase(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import TestCase

                testcase = cast(TestCase, node)
                if testcase.name is None:
                    return

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    r = range_from_node(testcase)
                    symbol = DocumentSymbol(
                        name=testcase.name, kind=SymbolKind.METHOD, range=r, selection_range=r, children=[]
                    )
                    self.current_symbol.children.append(symbol)

                    await self.generic_visit_current_symbol(node, symbol)

            async def visit_Keyword(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import Keyword

                keyword = cast(Keyword, node)
                if keyword.name is None:
                    return

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    r = range_from_node(keyword)
                    symbol = DocumentSymbol(
                        name=keyword.name, kind=SymbolKind.FUNCTION, range=r, selection_range=r, children=[]
                    )
                    self.current_symbol.children.append(symbol)

                    await self.generic_visit_current_symbol(node, symbol)

            async def visit_Arguments(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.lexer.tokens import Token as RobotToken
                from robot.parsing.model.statements import Arguments

                n = cast(Arguments, node)
                arguments = n.get_tokens(RobotToken.ARGUMENT)

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    for argument_token in (cast(RobotToken, e) for e in arguments):
                        argument = self.get_variable_token(argument_token)

                        if argument is not None:
                            r = range_from_token(argument)

                            symbol = DocumentSymbol(
                                name=argument.value, kind=SymbolKind.VARIABLE, range=r, selection_range=r
                            )
                            if symbol.name not in map(lambda v: v.name, self.current_symbol.children):
                                self.current_symbol.children.append(symbol)

            def get_variable_token(self, token: Token) -> Optional[Token]:
                from robot.parsing.lexer.tokens import Token as RobotToken

                return next(
                    (
                        v
                        for v in itertools.dropwhile(
                            lambda t: t.type in RobotToken.NON_DATA_TOKENS,
                            tokenize_variables(token, ignore_errors=True),
                        )
                        if v.type == RobotToken.VARIABLE
                    ),
                    None,
                )

            async def visit_KeywordCall(self, node: ast.AST) -> None:  # noqa: N802
                from robot.errors import VariableError
                from robot.parsing.lexer.tokens import Token as RobotToken
                from robot.parsing.model.statements import KeywordCall

                # TODO  analyse "Set Local/Global/Suite Variable"

                keyword_call = cast(KeywordCall, node)
                for assign_token in keyword_call.get_tokens(RobotToken.ASSIGN):
                    if assign_token is None:
                        continue

                    if self.current_symbol is not None and self.current_symbol.children is not None:
                        try:
                            variable_token = self.get_variable_token(assign_token)

                            if variable_token is not None:
                                r = range_from_token(variable_token)

                                symbol = DocumentSymbol(
                                    name=variable_token.value, kind=SymbolKind.VARIABLE, range=r, selection_range=r
                                )
                                if symbol.name not in map(lambda v: v.name, self.current_symbol.children):
                                    self.current_symbol.children.append(symbol)

                        except VariableError:
                            pass

            async def visit_ForHeader(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.lexer.tokens import Token as RobotToken
                from robot.parsing.model.statements import ForHeader

                n = cast(ForHeader, node)
                variables = n.get_tokens(RobotToken.VARIABLE)

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    for variable in variables:
                        variable_token = self.get_variable_token(variable)
                        if variable_token is not None:
                            r = range_from_token(variable_token)
                            symbol = DocumentSymbol(
                                name=variable_token.value, kind=SymbolKind.VARIABLE, range=r, selection_range=r
                            )
                            if symbol.name not in map(lambda v: v.name, self.current_symbol.children):
                                self.current_symbol.children.append(symbol)

            async def visit_KeywordName(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.lexer.tokens import Token as RobotToken
                from robot.parsing.model.statements import KeywordName

                n = cast(KeywordName, node)
                name_token = cast(Token, n.get_token(RobotToken.KEYWORD_NAME))
                if name_token is None:
                    return

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    for variable in filter(
                        lambda e: e.type == RobotToken.VARIABLE,
                        tokenize_variables(name_token, identifiers="$", ignore_errors=True),
                    ):
                        variable_token = self.get_variable_token(variable)
                        if variable_token is not None:
                            r = range_from_token(variable_token)
                            symbol = DocumentSymbol(
                                name=variable_token.value, kind=SymbolKind.VARIABLE, range=r, selection_range=r
                            )
                            if symbol.name not in map(lambda v: v.name, self.current_symbol.children):
                                self.current_symbol.children.append(symbol)

            async def visit_Variable(self, node: ast.AST) -> None:  # noqa: N802
                from robot.api.parsing import Token as RobotToken
                from robot.parsing.model.statements import Variable
                from robot.variables import search_variable

                variable = cast(Variable, node)

                name_token = variable.get_token(RobotToken.VARIABLE)
                name = name_token.value

                if name is not None:

                    match = search_variable(name, ignore_errors=True)
                    if not match.is_assign(allow_assign_mark=True):
                        return None

                    if name.endswith("="):
                        name = name[:-1].rstrip()

                if self.current_symbol is not None and self.current_symbol.children is not None:
                    r = range_from_node(variable)
                    symbol = DocumentSymbol(name=name, kind=SymbolKind.VARIABLE, range=r, selection_range=r)
                    self.current_symbol.children.append(symbol)

        return await Visitor.find_from(await self.parent.documents_cache.get_model(document), self)



from __future__ import annotations

from os import PathLike
from pathlib import Path
from typing import Any, Union


def path_is_relative_to(path: Union[Path, str, PathLike[Any]], other_path: Union[Path, str, PathLike[Any]]) -> bool:
    try:
        Path(path).relative_to(other_path)
        return True
    except ValueError:
        return False



from ..common.server import LanguageServerBase
from .protocol import RobotLanguageServerProtocol


class RobotLanguageServer(LanguageServerBase[RobotLanguageServerProtocol]):
    def create_protocol(self) -> RobotLanguageServerProtocol:
        return RobotLanguageServerProtocol(self)



from __future__ import annotations

import asyncio
import concurrent.futures
import contextvars
import functools
import inspect
import os
import threading
import weakref
from collections import deque
from concurrent.futures.thread import ThreadPoolExecutor
from types import TracebackType
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Deque,
    Generic,
    Iterator,
    List,
    MutableSet,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    runtime_checkable,
)

from ..utils.inspect import ensure_coroutine

__all__ = [
    "AsyncEventIterator",
    "AsyncEvent",
    "async_event",
    "AsyncTaskingEventIterator",
    "AsyncTaskingEvent",
    "async_tasking_event_iterator",
    "async_tasking_event",
    "AsyncThreadingEventIterator",
    "AsyncThreadingEvent",
    "async_threading_event_iterator",
    "async_threading_event",
    "check_canceled",
    "run_in_thread",
    "run_coroutine_in_thread",
    "Lock",
    "create_sub_task",
    "FutureInfo",
]

_T = TypeVar("_T")

_TResult = TypeVar("_TResult")
_TCallable = TypeVar("_TCallable", bound=Callable[..., Any])


class AsyncEventResultIteratorBase(Generic[_TCallable, _TResult]):
    def __init__(self) -> None:
        self._lock = threading.RLock()

        self._listeners: MutableSet[weakref.ref[Any]] = set()
        self._loop = asyncio.get_event_loop()

    def add(self, callback: _TCallable) -> None:
        def remove_listener(ref: Any) -> None:
            with self._lock:
                self._listeners.remove(ref)

        with self._lock:
            if inspect.ismethod(callback):
                self._listeners.add(weakref.WeakMethod(callback, remove_listener))
            else:
                self._listeners.add(weakref.ref(callback, remove_listener))

    def remove(self, callback: _TCallable) -> None:
        with self._lock:
            try:
                if inspect.ismethod(callback):
                    self._listeners.remove(weakref.WeakMethod(callback))
                else:
                    self._listeners.remove(weakref.ref(callback))
            except KeyError:
                pass

    def __contains__(self, obj: Any) -> bool:
        if inspect.ismethod(obj):
            return weakref.WeakMethod(obj) in self._listeners
        else:
            return weakref.ref(obj) in self._listeners

    def __len__(self) -> int:
        return len(self._listeners)

    def __iter__(self) -> Iterator[_TCallable]:
        for r in self._listeners:
            c = r()
            if c is not None:
                yield c

    async def __aiter__(self) -> AsyncIterator[_TCallable]:
        for r in self.__iter__():
            yield r

    async def _notify(
        self, *args: Any, callback_filter: Optional[Callable[[_TCallable], bool]] = None, **kwargs: Any
    ) -> AsyncIterator[_TResult]:

        for method in filter(
            lambda x: callback_filter(x) if callback_filter is not None else True,
            set(self),
        ):
            result = method(*args, **kwargs)
            if inspect.isawaitable(result):
                result = await result

            yield result


class AsyncEventIterator(AsyncEventResultIteratorBase[_TCallable, _TResult]):
    def __call__(self, *args: Any, **kwargs: Any) -> AsyncIterator[_TResult]:
        return self._notify(*args, **kwargs)


class AsyncEvent(AsyncEventResultIteratorBase[_TCallable, _TResult]):
    async def __call__(self, *args: Any, **kwargs: Any) -> List[_TResult]:
        return [a async for a in self._notify(*args, **kwargs)]


_TEvent = TypeVar("_TEvent")


class AsyncEventDescriptorBase(Generic[_TCallable, _TResult, _TEvent]):
    def __init__(
        self, _func: _TCallable, factory: Callable[..., _TEvent], *factory_args: Any, **factory_kwargs: Any
    ) -> None:
        self._func = _func
        self.__factory = factory
        self.__factory_args = factory_args
        self.__factory_kwargs = factory_kwargs
        self._owner: Optional[Any] = None
        self._owner_name: Optional[str] = None

    def __set_name__(self, owner: Any, name: str) -> None:
        self._owner = owner
        self._owner_name = name

    def __get__(self, obj: Any, objtype: Type[Any]) -> _TEvent:
        if obj is None:
            return self  # type: ignore

        name = f"__async_event_{self._func.__name__}__"
        if not hasattr(obj, name):
            setattr(obj, name, self.__factory(*self.__factory_args, **self.__factory_kwargs))

        return cast("_TEvent", getattr(obj, name))


class async_event_iterator(  # noqa: N801
    AsyncEventDescriptorBase[_TCallable, Any, AsyncEventIterator[_TCallable, Any]]
):
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(_func, AsyncEventIterator[_TCallable, _TResult])


class async_event(AsyncEventDescriptorBase[_TCallable, Any, AsyncEvent[_TCallable, Any]]):  # noqa: N801
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(_func, AsyncEvent[_TCallable, _TResult])


_F = TypeVar("_F", bound=Callable[..., Any])


def threaded(enabled: bool = True) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        setattr(func, "__threaded__", enabled)
        return func

    return decorator


@runtime_checkable
class HasThreaded(Protocol):
    __threaded__: bool


class AsyncTaskingEventResultIteratorBase(AsyncEventResultIteratorBase[_TCallable, _TResult]):
    def __init__(self, *, task_name_prefix: Optional[str] = None) -> None:
        super().__init__()
        self._task_name_prefix = task_name_prefix or type(self).__qualname__

    async def _notify(  # type: ignore
        self,
        *args: Any,
        result_callback: Optional[Callable[[Optional[_TResult], Optional[BaseException]], Any]] = None,
        return_exceptions: Optional[bool] = True,
        callback_filter: Optional[Callable[[_TCallable], bool]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[Union[_TResult, BaseException]]:
        def _done(f: asyncio.Future[_TResult]) -> None:
            if result_callback is not None:
                try:
                    result_callback(f.result(), f.exception())
                except (SystemExit, KeyboardInterrupt):
                    raise
                except BaseException as e:
                    result_callback(None, e)

        awaitables: List[asyncio.Future[_TResult]] = []
        for method in filter(
            lambda x: callback_filter(x) if callback_filter is not None else True,
            set(self),
        ):
            if method is not None:
                if isinstance(method, HasThreaded) and cast(HasThreaded, method).__threaded__:
                    future = run_coroutine_in_thread(ensure_coroutine(method), *args, **kwargs)
                else:
                    future = create_sub_task(ensure_coroutine(method)(*args, **kwargs))

                awaitables.append(future)

                if result_callback is not None:
                    future.add_done_callback(_done)

        for a in asyncio.as_completed(awaitables):
            try:
                yield await a
            except asyncio.CancelledError:
                for f in awaitables:
                    if not f.done():
                        f.cancel()
                        try:
                            yield await a
                        except (SystemExit, KeyboardInterrupt):
                            raise
                        except BaseException:
                            pass

                raise
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as e:
                if return_exceptions:
                    yield e
                else:
                    raise


class AsyncTaskingEventIterator(AsyncTaskingEventResultIteratorBase[_TCallable, _TResult]):
    def __call__(self, *args: Any, **kwargs: Any) -> AsyncIterator[Union[_TResult, BaseException]]:
        return self._notify(*args, **kwargs)


def _get_name_prefix(descriptor: AsyncEventDescriptorBase[Any, Any, Any]) -> str:
    if descriptor._owner is None:
        return type(descriptor).__qualname__

    return f"{descriptor._owner.__qualname__}.{descriptor._owner_name}"


class AsyncTaskingEvent(AsyncTaskingEventResultIteratorBase[_TCallable, _TResult]):
    async def __call__(self, *args: Any, **kwargs: Any) -> List[Union[_TResult, BaseException]]:
        return [a async for a in self._notify(*args, **kwargs)]


class AsyncThreadingEventResultIteratorBase(AsyncEventResultIteratorBase[_TCallable, _TResult]):
    __executor: Optional[ThreadPoolExecutor] = None

    def __init__(self, *, thread_name_prefix: Optional[str] = None) -> None:
        super().__init__()
        self.__executor = None
        self.__thread_name_prefix = thread_name_prefix or type(self).__qualname__

    def __del__(self) -> None:
        if self.__executor:
            self.__executor.shutdown(False)

    def _run_in_asyncio_thread(
        self,
        executor: ThreadPoolExecutor,
        coro: Union[asyncio.Future[_TResult], Awaitable[_TResult]],
        method_name: Optional[str] = None,
    ) -> asyncio.Future[_TResult]:
        def run(loop: asyncio.AbstractEventLoop) -> None:
            if method_name is not None:
                threading.current_thread().name = (
                    self.__thread_name_prefix() if callable(self.__thread_name_prefix) else self.__thread_name_prefix
                ) + f"->{method_name}(...)"

            asyncio.set_event_loop(loop)
            try:
                loop.run_forever()
            finally:
                loop.close()

        loop = asyncio.new_event_loop()

        # loop.set_debug(True)

        executor.submit(run, loop)

        result = asyncio.wrap_future(asyncio.run_coroutine_threadsafe(coro, loop=loop))

        def stop_loop(t: asyncio.Future[_TResult]) -> None:
            async def loop_stop() -> bool:
                loop.stop()
                return True

            asyncio.run_coroutine_threadsafe(loop_stop(), loop=loop)

        result.add_done_callback(stop_loop)
        return result

    async def _notify(  # type: ignore
        self,
        *args: Any,
        result_callback: Optional[Callable[[Optional[_TResult], Optional[BaseException]], Any]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        return_exceptions: Optional[bool] = True,
        callback_filter: Optional[Callable[[_TCallable], bool]] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[Union[_TResult, BaseException], None]:
        def _done(f: asyncio.Future[_TResult]) -> None:
            if result_callback is not None:
                try:
                    result_callback(f.result(), f.exception())
                except (SystemExit, KeyboardInterrupt):
                    raise
                except BaseException as e:
                    result_callback(None, e)

        if executor is None:
            if AsyncThreadingEventResultIteratorBase.__executor is None:
                AsyncThreadingEventResultIteratorBase.__executor = ThreadPoolExecutor(
                    thread_name_prefix=self.__thread_name_prefix()
                    if callable(self.__thread_name_prefix)
                    else self.__thread_name_prefix
                )
            executor = AsyncThreadingEventResultIteratorBase.__executor

        awaitables: List[asyncio.Future[_TResult]] = []
        for method in filter(
            lambda x: callback_filter(x) if callback_filter is not None else True,
            set(self),
        ):
            if method is not None:
                future = self._run_in_asyncio_thread(
                    executor,
                    ensure_coroutine(method)(*args, **kwargs),
                    method.__qualname__,
                )
                if result_callback is not None:
                    future.add_done_callback(_done)
                awaitables.append(future)

        for a in asyncio.as_completed(awaitables):
            try:
                yield await a
            except asyncio.CancelledError:
                for f in awaitables:
                    if not f.done():
                        f.cancel()
                        try:
                            yield await a
                        except (SystemExit, KeyboardInterrupt):
                            raise
                        except BaseException:
                            pass
                raise
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as e:
                if return_exceptions:
                    yield e
                else:
                    raise


class AsyncThreadingEventIterator(AsyncThreadingEventResultIteratorBase[_TCallable, _TResult]):
    def __call__(self, *args: Any, **kwargs: Any) -> AsyncIterator[Union[_TResult, BaseException]]:
        return self._notify(*args, **kwargs)


class async_threading_event_iterator(  # noqa: N801
    AsyncEventDescriptorBase[_TCallable, Any, AsyncThreadingEventIterator[_TCallable, Any]]
):
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(
            _func, AsyncThreadingEventIterator[_TCallable, Any], thread_name_prefix=lambda: _get_name_prefix(self)
        )


class AsyncThreadingEvent(AsyncThreadingEventResultIteratorBase[_TCallable, _TResult]):
    async def __call__(self, *args: Any, **kwargs: Any) -> List[Union[_TResult, BaseException]]:
        return [a async for a in self._notify(*args, **kwargs)]


class async_threading_event(  # noqa: N801
    AsyncEventDescriptorBase[_TCallable, Any, AsyncThreadingEvent[_TCallable, Any]]
):
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(_func, AsyncThreadingEvent[_TCallable, Any], thread_name_prefix=lambda: _get_name_prefix(self))


class async_tasking_event_iterator(  # noqa: N801
    AsyncEventDescriptorBase[_TCallable, Any, AsyncTaskingEventIterator[_TCallable, Any]]
):
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(
            _func, AsyncTaskingEventIterator[_TCallable, Any], task_name_prefix=lambda: _get_name_prefix(self)
        )


class async_tasking_event(AsyncEventDescriptorBase[_TCallable, Any, AsyncTaskingEvent[_TCallable, Any]]):  # noqa: N801
    def __init__(self, _func: _TCallable) -> None:
        super().__init__(_func, AsyncTaskingEvent[_TCallable, Any], task_name_prefix=lambda: _get_name_prefix(self))


async def check_canceled() -> bool:
    await asyncio.sleep(0)

    return True


def check_canceled_sync() -> bool:
    info = get_current_future_info()
    if info is not None and info.canceled():
        raise asyncio.CancelledError()
    return True


THREADPOOL_POOL_MAX_WORKERS = None

__tread_pool_executor: Optional[ThreadPoolExecutor] = None


def run_in_thread(func: Callable[..., _T], /, *args: Any, **kwargs: Any) -> asyncio.Future[_T]:
    global __tread_pool_executor
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)

    if __tread_pool_executor is None:
        __tread_pool_executor = ThreadPoolExecutor(
            max_workers=(
                int(s)
                if (s := os.environ.get("ROBOT_THREADPOOL_POOL_MAX_WORKERS", None)) and s.isnumeric()
                else THREADPOOL_POOL_MAX_WORKERS
            )
        )

    return cast("asyncio.Future[_T]", loop.run_in_executor(__tread_pool_executor, cast(Callable[..., _T], func_call)))


def run_coroutine_in_thread(
    coro: Callable[..., Coroutine[Any, Any, _T]], *args: Any, **kwargs: Any
) -> asyncio.Future[_T]:
    callback_added_event = Event()
    inner_task: Optional[asyncio.Task[_T]] = None
    canceled = False
    result: Optional[asyncio.Future[_T]] = None

    async def create_inner_task() -> _T:
        nonlocal inner_task

        ct = asyncio.current_task()

        old_name = threading.current_thread().getName()
        threading.current_thread().setName(coro.__qualname__)
        try:
            await callback_added_event.wait()

            if ct is not None and result is not None:
                _running_tasks[result].children.add(ct)

            inner_task = create_sub_task(coro(*args, **kwargs))

            if canceled:
                inner_task.cancel()

            return await inner_task
        finally:
            threading.current_thread().setName(old_name)

    def run() -> _T:
        return asyncio.run(create_inner_task())

    cti = get_current_future_info()
    result = run_in_thread(run)

    _running_tasks[result] = FutureInfo(result)
    if cti is not None:
        cti.children.add(result)

    def done(task: asyncio.Future[_T]) -> None:
        nonlocal canceled

        canceled = task.cancelled()

        if task.cancelled() and inner_task is not None and not inner_task.done():
            inner_task._loop.call_soon_threadsafe(inner_task.cancel)

    result.add_done_callback(done)

    callback_added_event.set()

    return result


class Event:
    """Thread safe version of an async Event"""

    def __init__(self) -> None:
        self._waiters: Deque[asyncio.Future[Any]] = deque()
        self._value = False
        self._lock = threading.RLock()

    def __repr__(self) -> str:
        res = super().__repr__()
        extra = "set" if self._value else "unset"
        if self._waiters:
            extra = f"{extra}, waiters:{len(self._waiters)}"
        return f"<{res[1:-1]} [{extra}]>"

    def is_set(self) -> bool:
        return self._value

    def set(self) -> None:
        if not self._value:
            self._value = True

            with self._lock:
                for fut in self._waiters:
                    if not fut.done():
                        if fut._loop == asyncio.get_running_loop():
                            fut.set_result(True)
                        else:
                            fut._loop.call_soon_threadsafe(fut.set_result, True)

    def clear(self) -> None:
        self._value = False

    async def wait(self) -> bool:
        if self._value:
            return True

        with self._lock:
            fut = create_sub_future()
            self._waiters.append(fut)
        try:
            await fut
            return True
        finally:
            with self._lock:
                self._waiters.remove(fut)


class Lock:
    """Threadsafe version of an async Lock."""

    def __init__(self) -> None:
        self._waiters: Optional[Deque[asyncio.Future[Any]]] = None
        self._locked = False
        self._lock = threading.RLock()

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self.release()

    def __repr__(self) -> str:
        res = super().__repr__()
        extra = "locked" if self._locked else "unlocked"
        if self._waiters:
            extra = f"{extra}, waiters:{len(self._waiters)}"
        return f"<{res[1:-1]} [{extra}]>"

    def locked(self) -> bool:
        return self._locked

    async def acquire(self) -> bool:

        with self._lock:
            if not self._locked and (self._waiters is None or all(w.cancelled() for w in self._waiters)):
                self._locked = True
                return True

        if self._waiters is None:
            self._waiters = deque()
        fut = create_sub_future()
        with self._lock:
            self._waiters.append(fut)

        try:
            try:
                await fut
            finally:
                self._waiters.remove(fut)
        except asyncio.CancelledError:
            if not self._locked:
                self._wake_up_first()
            raise

        with self._lock:
            self._locked = True

        return True

    def release(self) -> None:
        if self._locked:
            with self._lock:
                self._locked = False
            self._wake_up_first()
        else:
            raise RuntimeError("Lock is not acquired.")

    def _wake_up_first(self) -> None:
        if not self._waiters:
            return
        try:
            fut = next(iter(self._waiters))
        except StopIteration:
            return

        if not fut.done():
            if fut._loop == asyncio.get_running_loop():
                fut.set_result(True)
            else:
                fut._loop.call_soon_threadsafe(fut.set_result, True)


class FutureInfo:
    def __init__(self, future: asyncio.Future[Any]) -> None:
        self.task: weakref.ref[asyncio.Future[Any]] = weakref.ref(future)
        self.children: weakref.WeakSet[asyncio.Future[Any]] = weakref.WeakSet()

        future.add_done_callback(self._done)

    def _done(self, future: asyncio.Future[Any]) -> None:
        if future.cancelled():
            for t in self.children.copy():
                if not t.done() and not t.cancelled():

                    if t._loop == asyncio.get_running_loop():
                        t.cancel()
                    else:
                        t._loop.call_soon_threadsafe(t.cancel)

    def canceled(self) -> bool:
        task = self.task()
        if task is not None and task.cancelled():
            return True
        return False


_running_tasks: weakref.WeakKeyDictionary[asyncio.Future[Any], FutureInfo] = weakref.WeakKeyDictionary()


def get_current_future_info() -> Optional[FutureInfo]:
    ct = asyncio.current_task()

    if ct is None:
        return None

    if ct not in _running_tasks:
        _running_tasks[ct] = FutureInfo(ct)

    return _running_tasks[ct]


def create_sub_task(coro: Awaitable[_T], *, name: Optional[str] = None) -> asyncio.Task[_T]:

    ct = get_current_future_info()

    result = asyncio.create_task(coro, name=name)

    if ct is not None:
        ct.children.add(result)

    _running_tasks[result] = FutureInfo(result)
    return result


def create_sub_future(loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Future[Any]:

    ct = get_current_future_info()

    if loop is None:
        loop = asyncio.get_running_loop()

    result = loop.create_future()

    _running_tasks[result] = FutureInfo(result)

    if ct is not None:
        ct.children.add(result)

    return result


class _FutureHolder(Generic[_T]):
    def __init__(self, cfuture: concurrent.futures.Future[_T]):
        self.cfuture = cfuture
        self.afuture = wrap_sub_future(cfuture)


def spawn_coroutine_from_thread(
    func: Callable[..., Coroutine[Any, Any, _T]],
    *args: Any,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    **kwargs: Any,
) -> concurrent.futures.Future[_T]:
    if loop is None:
        loop = asyncio.get_running_loop()

    result = _FutureHolder(asyncio.run_coroutine_threadsafe(func(*args), loop))
    return result.cfuture


def run_coroutine_from_thread(
    func: Callable[..., Coroutine[Any, Any, _T]],
    *args: Any,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    **kwargs: Any,
) -> _T:
    if loop is None:
        loop = asyncio.get_running_loop()

    result = _FutureHolder(asyncio.run_coroutine_threadsafe(func(*args), loop))

    return result.cfuture.result()


def run_coroutine_from_thread_as_future_async(
    func: Callable[..., Coroutine[Any, Any, _T]],
    *args: Any,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    **kwargs: Any,
) -> asyncio.Future[_T]:
    if loop is None:
        loop = asyncio.get_running_loop()

    return wrap_sub_future(asyncio.run_coroutine_threadsafe(func(*args, **kwargs), loop))


async def run_coroutine_from_thread_async(
    func: Callable[..., Coroutine[Any, Any, _T]],
    *args: Any,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    **kwargs: Any,
) -> _T:
    if loop is None:
        loop = asyncio.get_running_loop()

    return await run_coroutine_from_thread_as_future_async(func, *args, loop=loop, **kwargs)


def wrap_sub_future(
    future: Union[asyncio.Future[_T], concurrent.futures.Future[_T]],
    *,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> asyncio.Future[_T]:
    result = asyncio.wrap_future(future, loop=loop)
    ci = get_current_future_info()
    if ci is not None:
        ci.children.add(result)
    return result



from __future__ import annotations

import atexit
import os
import signal
from concurrent.futures.process import ProcessPoolExecutor
from typing import Any, Optional

PROCESS_POOL_MAX_WORKERS = None

_process_pool: Optional[ProcessPoolExecutor] = None


def shutdown_process_pool() -> None:
    global _process_pool
    try:
        if _process_pool is not None:
            _process_pool.shutdown(True)
            _process_pool = None
    except (KeyboardInterrupt, SystemExit):
        raise
    except BaseException:  # NOSONAR
        pass


def _terminate(*args: Any, **kwargs: Any) -> None:
    shutdown_process_pool()


# we need this, because ProcessPoolExecutor is not correctly initialized if asyncio is reading from stdin
def get_process_pool() -> ProcessPoolExecutor:
    global _process_pool
    if _process_pool is None:
        _process_pool = ProcessPoolExecutor(
            max_workers=(
                int(s)
                if (s := os.environ.get("ROBOT_PROCESS_POOL_MAX_WORKERS", None)) and s.isnumeric()
                else PROCESS_POOL_MAX_WORKERS
            ),
            initializer=_init_pool,
        )
        atexit.register(_terminate)

        try:
            _process_pool.submit(_dummy_first_run_pool).result(5)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException:
            pass

    return _process_pool


def _init_pool() -> None:
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def _dummy_first_run_pool() -> None:
    """Dummy function to initialize the ProcessPoolExecutor"""
    pass



from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from enum import Enum, IntEnum, IntFlag
from reprlib import recursive_repr
from typing import Any, Dict, Iterator, List, Literal, Optional, Tuple, Union

from ...utils.dataclasses import to_camel_case, to_snake_case

ProgressToken = Union[str, int]
DocumentUri = str
URI = str


@dataclass(repr=False)
class Model:
    @classmethod
    def _encode_case(cls, s: str) -> str:
        return to_camel_case(s)

    @classmethod
    def _decode_case(cls, s: str) -> str:
        return to_snake_case(s)

    @recursive_repr()
    def __repr__(self) -> str:
        args = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}"
            for f in dataclasses.fields(self)
            if f.repr
            and (
                (f.default == dataclasses.MISSING and f.default_factory == dataclasses.MISSING)  # type: ignore
                or (f.default != dataclasses.MISSING and f.default != getattr(self, f.name))
                or (
                    f.default_factory != dataclasses.MISSING  # type: ignore
                    and getattr(self, f.name) != f.default_factory()
                )
            )
        )
        return f"{self.__class__.__qualname__}({args})"


@dataclass(repr=False)
class CancelParams(Model):
    id: Union[int, str]


@dataclass(repr=False)
class WorkDoneProgressParams(Model):
    work_done_token: Optional[ProgressToken] = None


@dataclass(repr=False)
class ClientInfo(Model):
    name: str
    version: Optional[str] = None


class TraceValue(Enum):
    OFF = "off"
    MESSAGE = "message"
    VERBOSE = "verbose"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class WorkspaceFolder(Model):
    uri: DocumentUri
    name: str


@dataclass(repr=False)
class TextDocumentSyncClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    will_save: Optional[bool] = None
    will_save_wait_until: Optional[bool] = None
    did_save: Optional[bool] = None


class ResourceOperationKind(Enum):
    CREATE = "create"
    RENAME = "rename"
    DELETE = "delete"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class FailureHandlingKind(Enum):
    ABORT = "abort"
    TRANSACTIONAL = "transactional"
    TEXT_ONLY_TRANSACTIONAL = "textOnlyTransactional"
    UNDO = "undo"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class WorkspaceEditClientCapabilitiesChangeAnnotationSupport(Model):
    groups_on_label: Optional[bool] = None


@dataclass(repr=False)
class WorkspaceEditClientCapabilities(Model):
    document_changes: Optional[bool] = None
    resource_operations: Optional[List[ResourceOperationKind]] = None
    failure_handling: Optional[FailureHandlingKind] = None
    normalizes_line_endings: Optional[bool] = None

    change_annotation_support: Optional[WorkspaceEditClientCapabilitiesChangeAnnotationSupport] = None


@dataclass(repr=False)
class DidChangeConfigurationClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DidChangeWatchedFilesClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class ExecuteCommandClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class SemanticTokensWorkspaceClientCapabilities(Model):
    refresh_support: Optional[bool] = None


@dataclass(repr=False)
class CodeLensWorkspaceClientCapabilities(Model):
    refresh_support: Optional[bool] = None


class SymbolKind(IntEnum):
    FILE = 1
    MODULE = 2
    NAMESPACE = 3
    PACKAGE = 4
    CLASS = 5
    METHOD = 6
    PROPERTY = 7
    FIELD = 8
    CONSTRUCTOR = 9
    ENUM = 10
    INTERFACE = 11
    FUNCTION = 12
    VARIABLE = 13
    CONSTANT = 14
    STRING = 15
    NUMBER = 16
    BOOLEAN = 17
    ARRAY = 18
    OBJECT = 19
    KEY = 20
    NULL = 21
    ENUMMEMBER = 22
    STRUCT = 23
    EVENT = 24
    OPERATOR = 25
    TYPEPARAMETER = 26

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class MarkupKind(Enum):
    PLAINTEXT = "plaintext"
    MARKDOWN = "markdown"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class CompletionItemTag(IntEnum):
    Deprecated = 1

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class SymbolTag(IntEnum):
    Deprecated = 1

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class InsertTextMode(IntEnum):
    AS_IS = 1
    ADJUST_INDENTATION = 2

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class InsertTextFormat(Enum):
    PLAINTEXT = 1
    SNIPPET = 2

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class WorkspaceSymbolClientCapabilitiesSymbolKind(Model):
    value_set: List[SymbolKind]


@dataclass(repr=False)
class WorkspaceSymbolClientCapabilitiesTagSupport(Model):
    value_set: List[SymbolTag]


@dataclass(repr=False)
class WorkspaceSymbolClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    symbol_kind: Optional[WorkspaceSymbolClientCapabilitiesSymbolKind] = None
    tag_support: Optional[WorkspaceSymbolClientCapabilitiesTagSupport] = None


class CompletionItemKind(IntEnum):
    TEXT = 1
    METHOD = 2
    FUNCTION = 3
    CONSTRUCTOR = 4
    FIELD = 5
    VARIABLE = 6
    CLASS = 7
    INTERFACE = 8
    MODULE = 9
    PROPERTY = 10
    UNIT = 11
    VALUE = 12
    ENUM = 13
    KEYWORD = 14
    SNIPPET = 15
    COLOR = 16
    FILE = 17
    REFERENCE = 18
    FOLDER = 19
    ENUM_MEMBER = 20
    CONSTANT = 21
    STRUCT = 22
    EVENT = 23
    OPERATOR = 24
    TYPE_PARAMETER = 25

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class CompletionClientCapabilitiesCompletionItemTagSupport(Model):
    value_set: List[CompletionItemTag]


@dataclass(repr=False)
class CompletionClientCapabilitiesCompletionItemResolveSupport(Model):
    properties: List[str]


@dataclass(repr=False)
class CompletionClientCapabilitiesCompletionItemInsertTextModeSupport(Model):
    value_set: List[InsertTextMode]


@dataclass(repr=False)
class CompletionClientCapabilitiesCompletionItem(Model):
    snippet_support: Optional[bool] = None
    commit_characters_support: Optional[bool] = None
    documentation_format: Optional[List[MarkupKind]] = None
    deprecated_support: Optional[bool] = None
    preselect_support: Optional[bool] = None
    tag_support: Optional[CompletionClientCapabilitiesCompletionItemTagSupport] = None
    insert_replace_support: Optional[bool] = None
    resolve_support: Optional[CompletionClientCapabilitiesCompletionItemResolveSupport] = None
    insert_text_mode_support: Optional[CompletionClientCapabilitiesCompletionItemInsertTextModeSupport] = None


@dataclass(repr=False)
class CompletionClientCapabilitiesCompletionItemKind(Model):
    value_set: Optional[List[CompletionItemKind]] = None


@dataclass(repr=False)
class CompletionClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    completion_item: Optional[CompletionClientCapabilitiesCompletionItem] = None
    completion_item_kind: Optional[CompletionClientCapabilitiesCompletionItemKind] = None
    context_support: Optional[bool] = None


@dataclass(repr=False)
class HoverClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    content_format: Optional[List[MarkupKind]] = None


@dataclass(repr=False)
class SignatureHelpClientCapabilitiesSignatureInformationParameterInformation(Model):
    label_offset_support: Optional[bool] = None


@dataclass(repr=False)
class SignatureHelpClientCapabilitiesSignatureInformation(Model):
    documentation_format: Optional[List[MarkupKind]] = None
    parameter_information: Optional[SignatureHelpClientCapabilitiesSignatureInformationParameterInformation] = None
    active_parameter_support: Optional[bool] = None


@dataclass(repr=False)
class SignatureHelpClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    signature_information: Optional[SignatureHelpClientCapabilitiesSignatureInformation] = None
    context_support: Optional[bool] = None


@dataclass(repr=False)
class DeclarationClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    link_support: Optional[bool] = None


@dataclass(repr=False)
class DefinitionClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    link_support: Optional[bool] = None


@dataclass(repr=False)
class TypeDefinitionClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    link_support: Optional[bool] = None


@dataclass(repr=False)
class ImplementationClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    link_support: Optional[bool] = None


@dataclass(repr=False)
class ReferenceClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentHighlightClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentSymbolClientCapabilitiesSymbolKind(Model):
    value_set: Optional[List[SymbolKind]] = None


@dataclass(repr=False)
class DocumentSymbolClientCapabilitiesTagSupport(Model):
    value_set: List[SymbolTag]


@dataclass(repr=False)
class DocumentSymbolClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    symbol_kind: Optional[DocumentSymbolClientCapabilitiesSymbolKind] = None
    hierarchical_document_symbol_support: Optional[bool] = None
    tag_support: Optional[DocumentSymbolClientCapabilitiesTagSupport] = None
    label_support: Optional[bool] = None


CodeActionKind = str


class CodeActionKinds(Enum):
    EMPTY = ""
    QUICKFIX = "quickfix"
    REFACTOR = "refactor"
    REFACTOREXTRACT = "refactor.extract"
    REFACTORINLINE = "refactor.inline"
    REFACTORREWRITE = "refactor.rewrite"
    SOURCE = "source"
    SOURCEORGANIZEIMPORTS = "source.organizeImports"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class CodeActionClientCapabilitiesCodeActionLiteralSupportCodeActionKind(Model):
    value_set: Optional[List[CodeActionKind]] = None


@dataclass(repr=False)
class CodeActionClientCapabilitiesCodeActionLiteralSupport(Model):
    code_action_kind: CodeActionClientCapabilitiesCodeActionLiteralSupportCodeActionKind


@dataclass(repr=False)
class CodeActionClientCapabilitiesResolveSupport(Model):
    properties: List[str]


@dataclass(repr=False)
class CodeActionClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    code_action_literal_support: Optional[CodeActionClientCapabilitiesCodeActionLiteralSupport] = None
    is_preferred_support: Optional[bool] = None
    disabled_support: Optional[bool] = None
    data_support: Optional[bool] = None
    resolve_support: Optional[CodeActionClientCapabilitiesResolveSupport] = None
    honors_change_annotations: Optional[bool] = None


@dataclass(repr=False)
class CodeLensClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentLinkClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    tooltip_support: Optional[bool] = None


@dataclass(repr=False)
class DocumentColorClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentFormattingClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentRangeFormattingClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class DocumentOnTypeFormattingClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class PrepareSupportDefaultBehavior(IntEnum):
    Identifier = 1

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class RenameClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    prepare_support: Optional[bool] = None
    prepare_support_default_behavior: Optional[PrepareSupportDefaultBehavior] = None
    honors_change_annotations: Optional[bool] = None


class DiagnosticTag(IntEnum):
    Unnecessary = 1
    Deprecated = 2

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class PublishDiagnosticsClientCapabilitiesTagSupport(Model):
    value_set: List[DiagnosticTag]


@dataclass(repr=False)
class PublishDiagnosticsClientCapabilities(Model):
    related_information: Optional[bool] = None
    tag_support: Optional[PublishDiagnosticsClientCapabilitiesTagSupport] = None
    version_support: Optional[bool] = None
    code_description_support: Optional[bool] = None
    data_support: Optional[bool] = None


@dataclass(repr=False)
class FoldingRangeClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    range_limit: Optional[int] = None
    line_folding_only: Optional[bool] = None


@dataclass(repr=False)
class SelectionRangeClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class LinkedEditingRangeClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class CallHierarchyClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


class TokenFormat(Enum):
    Relative = "relative"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class SemanticTokensClientCapabilitiesRequestsFull(Model):
    delta: Optional[bool] = None


@dataclass(repr=False)
class SemanticTokensClientCapabilitiesRequests(Model):
    range: Union[bool, Dict[Any, Any], None]
    full: Union[SemanticTokensClientCapabilitiesRequestsFull, bool, None] = None


@dataclass(repr=False)
class SemanticTokensClientCapabilities(Model):
    requests: SemanticTokensClientCapabilitiesRequests
    token_types: List[str]
    token_modifiers: List[str]
    formats: List[TokenFormat]
    overlapping_token_support: Optional[bool] = None
    multiline_token_support: Optional[bool] = None
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class MonikerClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None


@dataclass(repr=False)
class TextDocumentClientCapabilities(Model):
    synchronization: Optional[TextDocumentSyncClientCapabilities] = None
    completion: Optional[CompletionClientCapabilities] = None
    hover: Optional[HoverClientCapabilities] = None
    signature_help: Optional[SignatureHelpClientCapabilities] = None
    declaration: Optional[DeclarationClientCapabilities] = None
    definition: Optional[DefinitionClientCapabilities] = None
    type_definition: Optional[TypeDefinitionClientCapabilities] = None
    implementation: Optional[ImplementationClientCapabilities] = None
    references: Optional[ReferenceClientCapabilities] = None
    document_highlight: Optional[DocumentHighlightClientCapabilities] = None
    document_symbol: Optional[DocumentSymbolClientCapabilities] = None
    code_action: Optional[CodeActionClientCapabilities] = None
    code_lens: Optional[CodeLensClientCapabilities] = None
    document_link: Optional[DocumentLinkClientCapabilities] = None
    color_provider: Optional[DocumentColorClientCapabilities] = None
    formatting: Optional[DocumentFormattingClientCapabilities] = None
    range_formatting: Optional[DocumentRangeFormattingClientCapabilities] = None
    on_type_formatting: Optional[DocumentOnTypeFormattingClientCapabilities] = None
    rename: Optional[RenameClientCapabilities] = None
    publish_diagnostics: Optional[PublishDiagnosticsClientCapabilities] = None
    folding_range: Optional[FoldingRangeClientCapabilities] = None
    selection_range: Optional[SelectionRangeClientCapabilities] = None
    linked_editing_range: Optional[LinkedEditingRangeClientCapabilities] = None
    call_hierarchy: Optional[CallHierarchyClientCapabilities] = None
    semantic_tokens: Optional[SemanticTokensClientCapabilities] = None
    moniker: Optional[MonikerClientCapabilities] = None


@dataclass(repr=False)
class ShowMessageRequestClientCapabilitiesMessageActionItem(Model):
    additional_properties_support: Optional[bool] = None


@dataclass(repr=False)
class ShowMessageRequestClientCapabilities(Model):
    message_action_item: Optional[ShowMessageRequestClientCapabilitiesMessageActionItem] = None


@dataclass(repr=False)
class ShowDocumentClientCapabilities(Model):
    support: bool


@dataclass(repr=False)
class RegularExpressionsClientCapabilities(Model):
    engine: str
    version: Optional[str] = None


@dataclass(repr=False)
class MarkdownClientCapabilities(Model):
    parser: str
    version: Optional[str] = None


@dataclass(repr=False)
class ClientCapabilitiesWorkspaceFileOperationsWorkspaceClientCapabilities(Model):
    dynamic_registration: Optional[bool] = None
    did_create: Optional[bool] = None
    will_create: Optional[bool] = None
    did_rename: Optional[bool] = None
    will_rename: Optional[bool] = None
    did_delete: Optional[bool] = None
    will_delete: Optional[bool] = None


@dataclass(repr=False)
class ClientCapabilitiesWorkspace(Model):
    apply_edit: Optional[bool] = None
    workspace_edit: Optional[WorkspaceEditClientCapabilities] = None
    did_change_configuration: Optional[DidChangeConfigurationClientCapabilities] = None
    did_change_watched_files: Optional[DidChangeWatchedFilesClientCapabilities] = None
    symbol: Optional[WorkspaceSymbolClientCapabilities] = None
    execute_command: Optional[ExecuteCommandClientCapabilities] = None
    workspace_folders: Optional[bool] = None
    configuration: Optional[bool] = None
    semantic_tokens: Optional[SemanticTokensWorkspaceClientCapabilities] = None
    code_lens: Optional[CodeLensWorkspaceClientCapabilities] = None
    file_operations: Optional[ClientCapabilitiesWorkspaceFileOperationsWorkspaceClientCapabilities] = None


@dataclass(repr=False)
class ClientCapabilitiesWindow(Model):
    work_done_progress: Optional[bool] = None
    show_message: Optional[ShowMessageRequestClientCapabilities] = None
    show_document: Optional[ShowDocumentClientCapabilities] = None


@dataclass(repr=False)
class ClientCapabilitiesGeneral(Model):
    regular_expressions: Optional[RegularExpressionsClientCapabilities] = None
    markdown: Optional[MarkdownClientCapabilities] = None


@dataclass(repr=False)
class ClientCapabilities(Model):
    workspace: Optional[ClientCapabilitiesWorkspace] = None
    text_document: Optional[TextDocumentClientCapabilities] = None
    window: Optional[ClientCapabilitiesWindow] = None
    general: Optional[ClientCapabilitiesGeneral] = None
    experimental: Optional[Any] = None


@dataclass(repr=False)
class _InitializeParams(Model):
    capabilities: ClientCapabilities
    process_id: Optional[int] = None
    client_info: Optional[ClientInfo] = None
    locale: Optional[str] = None
    root_path: Optional[str] = None
    root_uri: Optional[DocumentUri] = None
    initialization_options: Optional[Any] = None
    trace: Optional[TraceValue] = None
    workspace_folders: Optional[List[WorkspaceFolder]] = None


@dataclass(repr=False)
class InitializeParams(WorkDoneProgressParams, _InitializeParams):
    pass


@dataclass(repr=False)
class InitializeError(Model):
    retry: bool


@dataclass(repr=False)
class WorkspaceFoldersServerCapabilities(Model):
    supported: Optional[bool] = None
    change_notifications: Union[str, bool, None] = None


class FileOperationPatternKind(Enum):
    FILE = "file"
    FOLDER = "folder"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class FileOperationPatternOptions(Model):
    ignore_case: Optional[bool] = None


@dataclass(repr=False)
class FileOperationPattern(Model):
    glob: str
    matches: Optional[FileOperationPatternKind] = None
    options: Optional[FileOperationPatternOptions] = None


@dataclass(repr=False)
class FileOperationFilter(Model):
    pattern: FileOperationPattern
    scheme: Optional[str] = None


@dataclass(repr=False)
class FileOperationRegistrationOptions(Model):
    filters: List[FileOperationFilter]


class TextDocumentSyncKind(Enum):
    NONE = 0
    FULL = 1
    INCREMENTAL = 2

    def __repr__(self) -> str:
        return super().__str__()


@dataclass(repr=False)
class SaveOptions(Model):
    include_text: Optional[bool]


@dataclass(repr=False)
class TextDocumentSyncOptions(Model):
    open_close: Optional[bool] = None
    change: Optional[TextDocumentSyncKind] = None
    will_save: Optional[bool] = None
    will_save_wait_until: Optional[bool] = None
    save: Union[bool, SaveOptions, None] = None


@dataclass(repr=False)
class WorkDoneProgressOptions(Model):
    work_done_progress: Optional[bool] = None


@dataclass(repr=False)
class DocumentFilter(Model):
    language: Optional[str] = None
    scheme: Optional[str] = None
    pattern: Optional[str] = None


DocumentSelector = List[DocumentFilter]


@dataclass(repr=False)
class TextDocumentRegistrationOptions(Model):
    document_selector: Optional[DocumentSelector] = None


@dataclass(repr=False)
class StaticRegistrationOptions(Model):
    id: Optional[str] = None


@dataclass(repr=False)
class TextDocumentChangeRegistrationOptions(TextDocumentRegistrationOptions):
    sync_kind: TextDocumentSyncKind = TextDocumentSyncKind.FULL


@dataclass(repr=False)
class FoldingRangeOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class FoldingRangeRegistrationOptions(StaticRegistrationOptions, TextDocumentRegistrationOptions, FoldingRangeOptions):
    pass


@dataclass(repr=False)
class DefinitionOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DeclarationOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DeclarationRegistrationOptions(DeclarationOptions, TextDocumentRegistrationOptions, StaticRegistrationOptions):
    pass


@dataclass(repr=False)
class ImplementationOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class ImplementationRegistrationOptions(DeclarationOptions, TextDocumentRegistrationOptions, StaticRegistrationOptions):
    pass


@dataclass(repr=False)
class HoverOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class WorkspaceSymbolOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DocumentSymbolOptions(WorkDoneProgressOptions):
    label: Optional[str] = None


@dataclass(repr=False)
class DocumentSymbolRegistrationOptions(TextDocumentRegistrationOptions, DocumentSymbolOptions):
    pass


@dataclass(repr=False)
class ServerCapabilitiesWorkspaceFileOperations(Model):
    did_create: Optional[FileOperationRegistrationOptions] = None
    will_create: Optional[FileOperationRegistrationOptions] = None
    did_rename: Optional[FileOperationRegistrationOptions] = None
    will_rename: Optional[FileOperationRegistrationOptions] = None
    did_delete: Optional[FileOperationRegistrationOptions] = None
    will_delete: Optional[FileOperationRegistrationOptions] = None


@dataclass(repr=False)
class ServerCapabilitiesWorkspace(Model):
    workspace_folders: Optional[WorkspaceFoldersServerCapabilities] = None
    file_operations: Optional[ServerCapabilitiesWorkspaceFileOperations] = None


@dataclass(repr=False)
class CompletionOptions(WorkDoneProgressOptions):
    trigger_characters: Optional[List[str]] = None
    all_commit_characters: Optional[List[str]] = None
    resolve_provider: Optional[bool] = None


@dataclass(repr=False)
class CompletionRegistrationOptions(TextDocumentRegistrationOptions, CompletionOptions):
    pass


@dataclass(repr=False)
class SignatureHelpOptions(WorkDoneProgressOptions):
    trigger_characters: Optional[List[str]] = None
    retrigger_characters: Optional[List[str]] = None


@dataclass(repr=False)
class CodeLensOptions(WorkDoneProgressOptions):
    resolve_provider: Optional[bool] = None


@dataclass(repr=False)
class DocumentFormattingOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DocumentFormattingRegistrationOptions(TextDocumentRegistrationOptions, DocumentFormattingOptions):
    pass


@dataclass(repr=False)
class DocumentRangeFormattingOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DocumentRangeFormattingRegistrationOptions(TextDocumentRegistrationOptions, DocumentRangeFormattingOptions):
    pass


@dataclass(repr=False)
class _ExecuteCommandOptions(Model):
    commands: List[str]


@dataclass(repr=False)
class ExecuteCommandOptions(WorkDoneProgressOptions, _ExecuteCommandOptions):
    pass


@dataclass(repr=False)
class SemanticTokensLegend(Model):
    token_types: List[str]
    token_modifiers: List[str]


@dataclass(repr=False)
class SemanticTokensOptionsFull(Model):
    delta: Optional[bool] = None


@dataclass(repr=False)
class SemanticTokensOptionsRange(Model):
    pass


@dataclass(repr=False)
class _SemanticTokensOptions(Model):
    legend: SemanticTokensLegend
    range: Union[bool, SemanticTokensOptionsRange, None] = None
    full: Union[bool, SemanticTokensOptionsFull, None] = None


@dataclass(repr=False)
class SemanticTokensOptions(WorkDoneProgressOptions, _SemanticTokensOptions):
    pass


@dataclass(repr=False)
class SemanticTokensRegistrationOptions(
    TextDocumentRegistrationOptions, StaticRegistrationOptions, SemanticTokensOptions
):
    pass


@dataclass(repr=False)
class ReferenceOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class ReferenceRegistrationOptions(TextDocumentRegistrationOptions, ReferenceOptions):
    pass


@dataclass(repr=False)
class DocumentHighlightOptions(WorkDoneProgressOptions):
    pass


@dataclass(repr=False)
class DocumentHighlightRegistrationOptions(TextDocumentRegistrationOptions, DocumentHighlightOptions):
    pass


@dataclass(repr=False)
class ServerCapabilities(Model):
    text_document_sync: Union[TextDocumentSyncOptions, TextDocumentSyncKind, None] = None
    completion_provider: Optional[CompletionOptions] = None
    hover_provider: Union[bool, HoverOptions, None] = None
    signature_help_provider: Optional[SignatureHelpOptions] = None
    declaration_provider: Union[bool, DeclarationOptions, DeclarationRegistrationOptions, None] = None
    definition_provider: Union[bool, DefinitionOptions, None] = None
    implementation_provider: Union[bool, ImplementationOptions, ImplementationRegistrationOptions, None] = None
    references_provider: Union[bool, ReferenceOptions, None] = None
    document_highlight_provider: Union[bool, DocumentHighlightOptions, None] = None
    document_symbol_provider: Union[bool, DocumentSymbolOptions, None] = None
    # TODO code_action_provider: Union[bool, CodeActionOptions] = None
    code_lens_provider: Optional[CodeLensOptions] = None
    # TODO document_link_provider: Optional[DocumentLinkOptions] = None
    # TODO color_provider: Union[bool, DocumentColorOptions, DocumentColorRegistrationOptions, None] = None
    document_formatting_provider: Union[bool, DocumentFormattingOptions, None] = None
    document_range_formatting_provider: Union[bool, DocumentRangeFormattingOptions, None] = None
    # TODO document_on_type_formatting_provider: Optional[DocumentOnTypeFormattingOptions] = None
    # TODO rename_provider: Union[bool, RenameOptions, None] = None
    folding_range_provider: Union[bool, FoldingRangeOptions, FoldingRangeRegistrationOptions, None] = None
    execute_command_provider: Optional[ExecuteCommandOptions] = None
    # TODO selection_range_provider: Union[bool, SelectionRangeOptions, SelectionRangeRegistrationOptions, None] = None
    # TODO linked_editing_range_provider: Union[
    #     boolean, LinkedEditingRangeOptions, LinkedEditingRangeRegistrationOptions, None
    # ] = None
    # TODO call_hierarchy_provider: Union[boolean, CallHierarchyOptions, CallHierarchyRegistrationOptions, None] = None
    semantic_tokens_provider: Union[SemanticTokensOptions, SemanticTokensRegistrationOptions, None] = None
    # TODO moniker_provider: Union[bool, MonikerOptions, MonikerRegistrationOptions, None] = None
    workspace_symbol_provider: Union[bool, WorkspaceSymbolOptions, None] = None
    workspace: Optional[ServerCapabilitiesWorkspace] = None
    experimental: Optional[Any] = None


@dataclass(repr=False)
class InitializeResultServerInfo(Model):
    name: str
    version: Optional[str] = None


@dataclass(repr=False)
class InitializeResult(Model):
    capabilities: ServerCapabilities
    server_info: Optional[InitializeResultServerInfo] = None


@dataclass(repr=False)
class InitializedParams(Model):
    pass


@dataclass(repr=False)
class DidChangeConfigurationParams(Model):
    settings: Any


@dataclass(repr=False)
class Position(Model):
    line: int
    character: int

    def __ge__(self, other: "Position") -> bool:
        line_gt = self.line > other.line

        if line_gt:
            return True

        if self.line == other.line:
            return self.character >= other.character

        return False

    def __gt__(self, other: "Position") -> bool:
        line_gt = self.line > other.line

        if line_gt:
            return True

        if self.line == other.line:
            return self.character > other.character

        return False

    def __le__(self, other: "Position") -> bool:
        line_lt = self.line < other.line

        if line_lt:
            return True

        if self.line == other.line:
            return self.character <= other.character

        return False

    def __lt__(self, other: "Position") -> bool:
        line_lt = self.line < other.line

        if line_lt:
            return True

        if self.line == other.line:
            return self.character < other.character

        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __iter__(self) -> Iterator[int]:
        return iter((self.line, self.character))

    def is_in_range(self, range: "Range") -> bool:
        return range.start <= self < range.end


@dataclass(repr=False)
class Range(Model):
    start: Position
    end: Position

    def __iter__(self) -> Iterator[Position]:
        return iter((self.start, self.end))

    @staticmethod
    def zero() -> "Range":
        return Range(
            start=Position(
                line=0,
                character=0,
            ),
            end=Position(
                line=0,
                character=0,
            ),
        )

    def extend(self, start_line: int = 0, start_character: int = 0, end_line: int = 0, end_character: int = 0) -> Range:
        return Range(
            start=Position(line=self.start.line + start_line, character=self.start.character + start_character),
            end=Position(line=self.end.line + end_line, character=self.end.character + end_character),
        )

    def __contains__(self, x: object) -> bool:
        if isinstance(x, (Position, Range)):
            return x.is_in_range(self)
        return False

    def is_in_range(self, range: Range) -> bool:
        return range.start.is_in_range(self) and range.end.is_in_range(self)


@dataclass(repr=False)
class TextDocumentItem(Model):
    uri: DocumentUri
    language_id: str
    version: int
    text: str


@dataclass(repr=False)
class DidOpenTextDocumentParams(Model):
    text_document: TextDocumentItem


@dataclass(repr=False)
class TextDocumentIdentifier(Model):
    uri: DocumentUri


@dataclass(repr=False)
class OptionalVersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: Optional[int] = None


@dataclass(repr=False)
class VersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: int


@dataclass(repr=False)
class DidCloseTextDocumentParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class TextDocumentContentRangeChangeEvent(Model):
    range: Range
    text: str
    range_length: Optional[int] = None


@dataclass(repr=False)
class TextDocumentContentTextChangeEvent(Model):
    text: str


TextDocumentContentChangeEvent = Union[TextDocumentContentRangeChangeEvent, TextDocumentContentTextChangeEvent]


@dataclass(repr=False)
class DidChangeTextDocumentParams(Model):
    text_document: VersionedTextDocumentIdentifier
    content_changes: List[TextDocumentContentChangeEvent]


@dataclass(repr=False)
class ConfigurationItem(Model):
    scope_uri: Optional[DocumentUri]
    section: Optional[str]


@dataclass(repr=False)
class ConfigurationParams(Model):
    items: List[ConfigurationItem]


class MessageType(IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    LOG = 4

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class ShowMessageParams(Model):
    type: MessageType
    message: str


@dataclass(repr=False)
class LogMessageParams(Model):
    type: MessageType
    message: str


@dataclass(repr=False)
class MessageActionItem(Model):
    title: str


@dataclass(repr=False)
class ShowMessageRequestParams(ShowMessageParams):
    actions: Optional[List[MessageActionItem]] = None


@dataclass(repr=False)
class ShowDocumentParams(Model):
    uri: URI
    external: Optional[bool] = None
    take_focus: Optional[bool] = None
    selection: Optional[Range] = None


@dataclass(repr=False)
class ShowDocumentResult(Model):
    success: bool


class TextDocumentSaveReason(IntEnum):
    Manual = 1
    AfterDelay = 2
    FocusOut = 3

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class WillSaveTextDocumentParams(Model):
    text_document: TextDocumentIdentifier
    reason: TextDocumentSaveReason


@dataclass(repr=False)
class TextEdit(Model):
    range: Range
    new_text: str


@dataclass(repr=False)
class DidSaveTextDocumentParams(Model):
    text_document: TextDocumentIdentifier
    text: Optional[str] = None


class DiagnosticSeverity(Enum):
    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    HINT = 4

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class CodeDescription(Model):
    href: URI


@dataclass(repr=False)
class Location(Model):
    uri: DocumentUri
    range: Range


@dataclass(repr=False)
class LocationLink(Model):
    origin_selection_range: Optional[Range]
    target_uri: DocumentUri
    target_range: Range
    target_selection_range: Range


@dataclass(repr=False)
class DiagnosticRelatedInformation(Model):
    location: Location
    message: str


@dataclass(repr=False)
class Diagnostic(Model):
    range: Range
    message: str
    severity: Optional[DiagnosticSeverity] = None
    code: Union[int, str, None] = None
    code_description: Optional[CodeDescription] = None
    source: Optional[str] = None
    tags: Optional[List[DiagnosticTag]] = None
    related_information: Optional[List[DiagnosticRelatedInformation]] = None
    data: Optional[Any] = None


@dataclass(repr=False)
class PublishDiagnosticsParams(Model):
    uri: DocumentUri
    diagnostics: List[Diagnostic]
    version: Optional[int] = None


@dataclass(repr=False)
class SetTraceParams(Model):
    value: TraceValue


@dataclass(repr=False)
class _FoldingRangeParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class FoldingRangeParams(WorkDoneProgressParams, _FoldingRangeParams):
    pass


class FoldingRangeKind(Enum):
    Comment = "comment"
    Imports = "imports"
    Region = "region"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class FoldingRange(Model):
    start_line: int
    end_line: int
    start_character: Optional[int] = None
    end_character: Optional[int] = None
    kind: Union[FoldingRangeKind, str, None] = None


@dataclass(repr=False)
class FileCreate(Model):
    uri: str


@dataclass(repr=False)
class CreateFilesParams(Model):
    files: List[FileCreate]


@dataclass(repr=False)
class FileRename(Model):
    old_uri: str
    new_uri: str


@dataclass(repr=False)
class RenameFilesParams(Model):
    files: List[FileRename]


@dataclass(repr=False)
class FileDelete(Model):
    uri: str


@dataclass(repr=False)
class DeleteFilesParams(Model):
    files: List[FileDelete]


ChangeAnnotationIdentifier = str


@dataclass(repr=False)
class CreateFileOptions(Model):
    overwrite: Optional[bool] = None
    ignore_if_exists: Optional[bool] = None


@dataclass(repr=False)
class CreateFile(Model):
    kind: Literal["create"]
    uri: DocumentUri
    options: Optional[CreateFileOptions]
    annotation_id: ChangeAnnotationIdentifier


@dataclass(repr=False)
class RenameFileOptions(Model):
    overwrite: Optional[bool] = None
    ignore_if_exists: Optional[bool] = None


@dataclass(repr=False)
class RenameFile(Model):
    kind: Literal["rename"]
    old_uri: DocumentUri
    new_uri: DocumentUri
    options: Optional[RenameFileOptions]
    annotation_id: ChangeAnnotationIdentifier


@dataclass(repr=False)
class DeleteFileOptions(Model):
    recursive: Optional[bool] = None
    ignore_if_exists: Optional[bool] = None


@dataclass(repr=False)
class DeleteFile(Model):
    kind: Literal["delete"]
    uri: DocumentUri
    options: Optional[DeleteFileOptions]
    annotation_id: ChangeAnnotationIdentifier


@dataclass(repr=False)
class AnnotatedTextEdit(TextEdit):
    annotation_id: ChangeAnnotationIdentifier


@dataclass(repr=False)
class TextDocumentEdit(Model):
    text_document: OptionalVersionedTextDocumentIdentifier
    edits: Union[TextEdit, AnnotatedTextEdit]


@dataclass(repr=False)
class ChangeAnnotation(Model):
    label: str
    needs_confirmation: Optional[bool] = None
    description: Optional[str] = None


@dataclass(repr=False)
class WorkspaceEdit(Model):
    changes: Optional[Dict[DocumentUri, List[TextEdit]]] = None
    document_changes: Union[List[TextDocumentEdit], TextDocumentEdit, CreateFile, RenameFile, DeleteFile, None] = None
    change_annotations: Optional[Dict[ChangeAnnotationIdentifier, ChangeAnnotation]] = None


@dataclass(repr=False)
class PartialResultParams(Model):
    partial_result_token: Optional[ProgressToken] = None


@dataclass(repr=False)
class TextDocumentPositionParams(Model):
    text_document: TextDocumentIdentifier
    position: Position


@dataclass(repr=False)
class DefinitionParams(WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams):
    pass


@dataclass(repr=False)
class DeclarationParams(WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams):
    pass


@dataclass(repr=False)
class ImplementationParams(WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams):
    pass


@dataclass(repr=False)
class HoverParams(WorkDoneProgressParams, TextDocumentPositionParams):
    pass


@dataclass(repr=False)
class MarkedStringType(Model):
    language: str
    value: str


MarkedString = Union[str, MarkedStringType]


@dataclass(repr=False)
class MarkupContent(Model):
    kind: MarkupKind
    value: str


@dataclass(repr=False)
class Hover(Model):
    contents: Union[MarkedString, List[MarkedString], MarkupContent]
    range: Optional[Range] = None


@dataclass(repr=False)
class WorkspaceFoldersChangeEvent(Model):
    added: List[WorkspaceFolder]
    removed: List[WorkspaceFolder]


@dataclass(repr=False)
class DidChangeWorkspaceFoldersParams(Model):
    event: WorkspaceFoldersChangeEvent


@dataclass(repr=False)
class Registration(Model):
    id: str
    method: str
    register_options: Optional[Any]


@dataclass(repr=False)
class RegistrationParams(Model):
    registrations: List[Registration]


@dataclass(repr=False)
class Unregistration(Model):
    id: str
    method: str


@dataclass(repr=False)
class UnregistrationParams(Model):
    unregisterations: List[Unregistration]


class WatchKind(IntFlag):
    CREATE = 1
    CHANGE = 2
    DELETE = 4


@dataclass(repr=False)
class FileSystemWatcher(Model):
    glob_pattern: str
    kind: Optional[WatchKind]


@dataclass(repr=False)
class DidChangeWatchedFilesRegistrationOptions(Model):
    watchers: List[FileSystemWatcher]


class FileChangeType(IntEnum):
    CREATED = 1
    CHANGED = 2
    DELETED = 3

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class FileEvent(Model):
    uri: DocumentUri
    type: FileChangeType


@dataclass(repr=False)
class DidChangeWatchedFilesParams(Model):
    changes: List[FileEvent]


@dataclass(repr=False)
class Command(Model):
    title: str
    command: str
    arguments: Optional[List[Any]] = None


class CompletionTriggerKind(Enum):
    INVOKED = 1
    TRIGGERCHARACTER = 2
    TRIGGERFORINCOMPLETECOMPLETIONS = 3

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class CompletionContext(Model):
    trigger_kind: CompletionTriggerKind
    trigger_character: Optional[str] = None


@dataclass(repr=False)
class CompletionParams(WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams):
    context: Optional[CompletionContext] = None


@dataclass(repr=False)
class InsertReplaceEdit(Model):
    new_text: str
    insert: Range
    replace: Range


@dataclass(repr=False)
class CompletionItem(Model):
    label: str
    kind: Optional[CompletionItemKind] = None
    tags: Optional[List[CompletionItemTag]] = None
    detail: Optional[str] = None
    documentation: Union[str, MarkupContent, None] = None
    deprecated: Optional[bool] = None
    preselect: Optional[bool] = None
    sort_text: Optional[str] = None
    filter_text: Optional[str] = None
    insert_text: Optional[str] = None
    insert_text_format: Optional[InsertTextFormat] = None
    insert_text_mode: Optional[InsertTextMode] = None
    text_edit: Union[TextEdit, InsertReplaceEdit, None] = None
    additional_text_edits: Optional[List[TextEdit]] = None
    commit_characters: Optional[List[str]] = None
    command: Optional[Command] = None
    data: Optional[Any] = None


@dataclass(repr=False)
class CompletionList(Model):
    is_incomplete: bool
    items: List[CompletionItem]


class SignatureHelpTriggerKind(Enum):
    INVOKED = 1
    TRIGGERCHARACTER = 2
    CONTENTCHANGE = 3

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class ParameterInformation(Model):
    label: Union[str, Tuple[int, int]]
    documentation: Union[str, MarkupContent, None] = None


@dataclass(repr=False)
class SignatureInformation(Model):
    label: str
    documentation: Union[str, MarkupContent, None] = None
    parameters: Optional[List[ParameterInformation]] = None
    active_parameter: Optional[int] = None


@dataclass(repr=False)
class SignatureHelp(Model):
    signatures: List[SignatureInformation]
    active_signature: Optional[int] = None
    active_parameter: Optional[int] = None


@dataclass(repr=False)
class SignatureHelpContext(Model):
    trigger_kind: SignatureHelpTriggerKind
    trigger_character: Optional[str] = None
    is_retrigger: bool = False
    active_signature_help: Optional[SignatureHelp] = None


@dataclass(repr=False)
class SignatureHelpParams(WorkDoneProgressParams, TextDocumentPositionParams):
    context: Optional[SignatureHelpContext] = None


@dataclass(repr=False)
class _CodeLensParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class CodeLensParams(WorkDoneProgressParams, PartialResultParams, _CodeLensParams):
    pass


@dataclass(repr=False)
class CodeLens(Model):
    range: Range
    command: Optional[Command] = None
    data: Optional[Any] = None


@dataclass(repr=False)
class CodeLensRegistrationOptions(TextDocumentRegistrationOptions, CodeLensOptions):
    pass


@dataclass(repr=False)
class _DocumentSymbolParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class DocumentSymbolParams(WorkDoneProgressParams, PartialResultParams, _DocumentSymbolParams):
    pass


@dataclass(repr=False)
class DocumentSymbol(Model):
    name: str
    kind: SymbolKind
    range: Range
    selection_range: Range
    detail: Optional[str] = None
    tags: Optional[List[SymbolTag]] = None
    deprecated: Optional[bool] = None
    children: Optional[List[DocumentSymbol]] = None


@dataclass(repr=False)
class SymbolInformation(Model):
    name: str
    kind: SymbolKind
    location: Location
    tags: Optional[List[SymbolTag]] = None
    deprecated: Optional[bool] = None
    container_name: Optional[str] = None


@dataclass(repr=False)
class FormattingOptions(Model):
    tab_size: int
    insert_spaces: bool
    trim_trailing_whitespace: Optional[bool] = None
    insert_final_newline: Optional[bool] = None
    trim_final_newlines: Optional[bool] = None


@dataclass(repr=False)
class _DocumentFormattingParams(Model):
    text_document: TextDocumentIdentifier
    options: FormattingOptions


@dataclass(repr=False)
class DocumentFormattingParams(WorkDoneProgressParams, _DocumentFormattingParams):
    pass


@dataclass(repr=False)
class _DocumentRangeFormattingParams(Model):
    text_document: TextDocumentIdentifier
    range: Range
    options: FormattingOptions


@dataclass(repr=False)
class DocumentRangeFormattingParams(WorkDoneProgressParams, _DocumentRangeFormattingParams):
    pass


@dataclass(repr=False)
class _SemanticTokensParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class SemanticTokensParams(WorkDoneProgressParams, PartialResultParams, _SemanticTokensParams):
    pass


@dataclass(repr=False)
class SemanticTokens(Model):
    data: List[int]
    result_id: Optional[str] = None


@dataclass(repr=False)
class SemanticTokensPartialResult(Model):
    data: List[int]


@dataclass(repr=False)
class _SemanticTokensDeltaParams(Model):
    text_document: TextDocumentIdentifier
    previous_result_id: str


@dataclass(repr=False)
class SemanticTokensDeltaParams(WorkDoneProgressParams, PartialResultParams, _SemanticTokensDeltaParams):
    pass


@dataclass(repr=False)
class SemanticTokensEdit(Model):
    start: int
    delete_count: int
    data: Optional[List[int]] = None


@dataclass(repr=False)
class SemanticTokensDelta(Model):
    edits: List[SemanticTokensEdit]
    result_id: Optional[str] = None


@dataclass(repr=False)
class SemanticTokensDeltaPartialResult(Model):
    edits: List[SemanticTokensEdit]


@dataclass(repr=False)
class _SemanticTokensRangeParams(Model):
    text_document: TextDocumentIdentifier
    range: Range


@dataclass(repr=False)
class SemanticTokensRangeParams(WorkDoneProgressParams, PartialResultParams, _SemanticTokensRangeParams):
    pass


class SemanticTokenTypes(Enum):
    NAMESPACE = "namespace"
    TYPE = "type"
    CLASS = "class"
    ENUM = "enum"
    INTERFACE = "interface"
    STRUCT = "struct"
    TYPE_PARAMETER = "typeParameter"
    PARAMETER = "parameter"
    VARIABLE = "variable"
    PROPERTY = "property"
    ENUM_MEMBER = "enumMember"
    EVENT = "event"
    FUNCTION = "function"
    METHOD = "method"
    MACRO = "macro"
    KEYWORD = "keyword"
    MODIFIER = "modifier"
    COMMENT = "comment"
    STRING = "string"
    NUMBER = "number"
    REGEXP = "regexp"
    OPERATOR = "operator"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class SemanticTokenModifiers(Enum):
    DECLARATION = "declaration"
    DEFINITION = "definition"
    READONLY = "readonly"
    STATIC = "static"
    DEPRECATED = "deprecated"
    ABSTRACT = "abstract"
    ASYNC = "async"
    MODIFICATION = "modification"
    DOCUMENTATION = "documentation"
    DEFAULT_LIBRARY = "defaultLibrary"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass(repr=False)
class ReferenceContext(Model):
    include_declaration: bool


@dataclass(repr=False)
class _ReferenceParams(Model):
    context: ReferenceContext


@dataclass(repr=False)
class ReferenceParams(WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams, _ReferenceParams):
    pass


@dataclass(repr=False)
class _DocumentHighlightParams(Model):
    pass


@dataclass(repr=False)
class DocumentHighlightParams(
    WorkDoneProgressParams, PartialResultParams, TextDocumentPositionParams, _DocumentHighlightParams
):
    pass


class DocumentHighlightKind(Enum):
    TEXT = 1
    READ = 2
    WRITE = 3


@dataclass(repr=False)
class DocumentHighlight(Model):
    range: Range
    kind: Optional[DocumentHighlightKind] = None
